"""Core building blocks for pages and databases."""

from __future__ import annotations

from datetime import datetime
from typing import TypeVar
from uuid import UUID

from notion_client.helpers import get_url

from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.objects import User
from ultimate_notion.utils import Wrapper, get_active_session

# Todo: Move the functionality from the PyDantic types in here, i.e. the currenctly commented code
T = TypeVar('T', bound=obj_blocks.DataObject)


class DataObject(Wrapper[T], wraps=obj_blocks.DataObject):
    """The base type for all data-related types, i.e, pages, databases and blocks."""

    def __eq__(self, other: object) -> bool:
        if other is None:
            return False
        elif not isinstance(other, DataObject):
            msg = f'Cannot compare {self.__class__.__name__} with {type(other).__name__}'
            raise RuntimeError(msg)
        else:
            return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def id(self) -> UUID:  # noqa: A003
        return self.obj_ref.id

    @property
    def created_time(self) -> datetime:
        return self.obj_ref.created_time

    @property
    def created_by(self) -> User:
        session = get_active_session()
        return session.get_user(self.obj_ref.created_by.id)

    @property
    def last_edited_time(self) -> datetime:
        return self.obj_ref.last_edited_time

    @property
    def last_edited_by(self) -> User:
        session = get_active_session()
        return session.get_user(self.obj_ref.last_edited_by.id)

    @property
    def parent(self) -> DataObject | None:
        """Return the parent record or None if the workspace is the parent."""
        session = get_active_session()
        parent = self.obj_ref.parent

        if isinstance(parent, objs.WorkspaceRef):
            return None
        elif isinstance(parent, objs.PageRef):
            return session.get_page(page_ref=parent.page_id)
        elif isinstance(parent, objs.DatabaseRef):
            return session.get_db(db_ref=parent.database_id)
        elif isinstance(parent, objs.BlockRef):
            return session._get_block(block_ref=parent.block_id)
        else:
            msg = f'Unknown parent reference {type(parent)}'
            raise RuntimeError(msg)

    @property
    def ancestors(self) -> tuple[DataObject, ...]:
        """Return all ancestors from the workspace to the actual record (excluding)."""
        match parent := self.parent:
            case None:
                return ()
            case _:
                return (*parent.ancestors, parent)

    @property
    def has_children(self) -> bool:
        return self.obj_ref.has_children

    @property
    def is_deleted(self) -> bool:
        """Return wether the object is deleted/archived."""
        return self.obj_ref.archived

    @property
    def url(self) -> str:
        return get_url(str(self.id))


class Block(DataObject[obj_blocks.Block], wraps=obj_blocks.Block):
    """General Notion block."""

    # ToDo: Implement me!


class TextBlock(DataObject[obj_blocks.TextBlock], wraps=obj_blocks.TextBlock):
    """Text block."""

    # ToDo: Implement me!


class ChildrenMixin:
    """Mixin for blocks that support children blocks."""

    # ToDo: Implement me!

    # @property
    # def __children__(self):
    #     """Provide short-hand access to the children in this block."""

    #     return self("children")

    # def __iadd__(self, block):
    #     """Append the given block to the children of this parent in place."""
    #     self.append(block)
    #     return self

    # def append(self, block):
    #     """Append the given block to the children of this parent."""

    #     if block is None:
    #         raise ValueError("block cannot be None")

    #     nested = self()

    #     if nested.children is None:
    #         nested.children = []

    #     nested.children.append(block)

    #     self.has_children = True


class Paragraph(DataObject[obj_blocks.Paragraph], ChildrenMixin, wraps=obj_blocks.Paragraph):
    """Paragraph block."""

    # ToDo: Implement me!


class Heading1(DataObject[obj_blocks.Heading1], wraps=obj_blocks.Heading1):
    """Heading 1 block."""

    # ToDo: Implement me!


class Heading2(DataObject[obj_blocks.Heading2], wraps=obj_blocks.Heading2):
    """Heading 2 block."""

    # ToDo: Implement me!


class Heading3(DataObject[obj_blocks.Heading3], wraps=obj_blocks.Heading3):
    """Heading 3 block."""

    # ToDo: Implement me!


class Quote(DataObject[obj_blocks.Quote], ChildrenMixin, wraps=obj_blocks.Quote):
    """Quote block."""

    # ToDo: Implement me!


class Code(DataObject[obj_blocks.Code], wraps=obj_blocks.Code):
    """Code block."""

    # ToDo: Implement me!


class Callout(DataObject[obj_blocks.Callout], wraps=obj_blocks.Callout):
    """Callout block."""

    # ToDo: Implement me!


class BulletedItem(DataObject[obj_blocks.BulletedListItem], ChildrenMixin, wraps=obj_blocks.BulletedListItem):
    """Bulleted list item."""

    # ToDo: Implement me!


class NumberedItem(DataObject[obj_blocks.NumberedListItem], ChildrenMixin, wraps=obj_blocks.NumberedListItem):
    """Numbered list item."""

    # ToDo: Implement me!


class ToDoItem(DataObject[obj_blocks.ToDo], ChildrenMixin, wraps=obj_blocks.ToDo):
    """ToDo list item."""

    # ToDo: Implement me!


class ToggleItem(DataObject[obj_blocks.Toggle], ChildrenMixin, wraps=obj_blocks.Toggle):
    """Toggle list item."""

    # ToDo: Implement me!


class Divider(DataObject[obj_blocks.Divider], wraps=obj_blocks.Divider):
    """Divider block."""

    # ToDo: Implement me!


class TableOfContents(DataObject[obj_blocks.TableOfContents], wraps=obj_blocks.TableOfContents):
    """Table of contents block."""

    # ToDo: Implement me!


class Breadcrumb(DataObject[obj_blocks.Breadcrumb], wraps=obj_blocks.Breadcrumb):
    """Breadcrumb block."""

    # ToDo: Implement me!


class Embed(DataObject[obj_blocks.Embed], wraps=obj_blocks.Embed):
    """Embed block."""

    # ToDo: Implement me!


class Bookmark(DataObject[obj_blocks.Bookmark], wraps=obj_blocks.Bookmark):
    """Bookmark block."""

    # ToDo: Implement me!


class LinkPreview(DataObject[obj_blocks.LinkPreview], wraps=obj_blocks.LinkPreview):
    """Link preview block."""

    # ToDo: Implement me!


class Equation(DataObject[obj_blocks.Equation], wraps=obj_blocks.Equation):
    """Equation block."""

    # ToDo: Implement me!


class File(DataObject[obj_blocks.File], wraps=obj_blocks.File):
    """File block."""

    # ToDo: Implement me!


class Image(DataObject[obj_blocks.Image], wraps=obj_blocks.Image):
    """Image block."""

    # ToDo: Implement me!


class Video(DataObject[obj_blocks.Video], wraps=obj_blocks.Video):
    """Video block."""

    # ToDo: Implement me!


class PDF(DataObject[obj_blocks.PDF], wraps=obj_blocks.PDF):
    """PDF block."""

    # ToDo: Implement me!


class ChildPage(DataObject[obj_blocks.ChildPage], wraps=obj_blocks.ChildPage):
    """Child page block."""

    # ToDo: Implement me!


class ChildDatabase(DataObject[obj_blocks.ChildDatabase], wraps=obj_blocks.ChildDatabase):
    """Child database block."""

    # ToDo: Implement me!


class Column(DataObject[obj_blocks.Column], ChildrenMixin, wraps=obj_blocks.Column):
    """Collumn block."""

    # ToDo: Implement me!


class ColumnList(DataObject[obj_blocks.ColumnList], ChildrenMixin, wraps=obj_blocks.ColumnList):
    """Column list block."""

    # ToDo: Implement me!


class TableRow(DataObject[obj_blocks.TableRow], wraps=obj_blocks.TableRow):
    """Table row block."""

    # ToDo: Implement me!


class Table(DataObject[obj_blocks.Table], ChildrenMixin, wraps=obj_blocks.Table):
    """Table block."""

    # ToDo: Implement me!


class LinkToPage(DataObject[obj_blocks.LinkToPage], wraps=obj_blocks.LinkToPage):
    """Link to page block."""

    # ToDo: Implement me!


class SyncedBlock(DataObject[obj_blocks.SyncedBlock], ChildrenMixin, wraps=obj_blocks.SyncedBlock):
    """Synced block - either original or synched."""

    @property
    def is_original(self) -> bool:
        """Is this block the original content."""
        return self.obj_ref.synced_block.synced_from is None

    @property
    def is_synched(self) -> bool:
        return not self.is_original


class Template(DataObject[obj_blocks.Template], ChildrenMixin, wraps=obj_blocks.Template):
    """Template block."""



"""Provides an object-based Notion API with all endpoints.

This pydantic based API is often referred to as just `api` while the low-level
API of the [Notion Client SDK library](https://github.com/ramnes/notion-sdk-py)
is just referred to as `raw_api`.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast
from uuid import UUID

from pydantic import SerializeAsAny, TypeAdapter

from ultimate_notion.obj_api.blocks import Block, Database, Page
from ultimate_notion.obj_api.iterator import EndpointIterator, PropertyItemList
from ultimate_notion.obj_api.objects import (
    DatabaseRef,
    EmojiObject,
    FileObject,
    ObjectReference,
    PageRef,
    ParentRef,
    RichTextObject,
    User,
)
from ultimate_notion.obj_api.props import PropertyItem, Title
from ultimate_notion.obj_api.query import DBQueryBuilder, SearchQueryBuilder
from ultimate_notion.obj_api.schema import PropertyType

if TYPE_CHECKING:
    from notion_client import Client as NCClient
    from notion_client.api_endpoints import BlocksChildrenEndpoint as NCBlocksChildrenEndpoint
    from notion_client.api_endpoints import BlocksEndpoint as NCBlocksEndpoint
    from notion_client.api_endpoints import DatabasesEndpoint as NCDatabasesEndpoint

logger = logging.getLogger(__name__)
T_UNSET: TypeAlias = Literal['UNSET']
UNSET: T_UNSET = 'UNSET'


class SessionError(Exception):
    """Raised when there are issues with the Notion session."""

    def __init__(self, message):
        """Initialize the `SessionError` with a supplied message."""
        super().__init__(message)


class NotionAPI:
    """Object-based Notion API (pydantic) with all endpoints."""

    def __init__(self, client: NCClient):
        self.client = client

        self.blocks = BlocksEndpoint(self)
        self.databases = DatabasesEndpoint(self)
        self.pages = PagesEndpoint(self)
        self.search = SearchEndpoint(self)
        self.users = UsersEndpoint(self)


class Endpoint:
    """Baseclass of the Notion API endpoints."""

    def __init__(self, api: NotionAPI):
        self.api = api


class BlocksEndpoint(Endpoint):
    """Interface to the 'blocks' endpoint of the Notion API"""

    class ChildrenEndpoint(Endpoint):
        """Interface to the API 'blocks/children' endpoint."""

        @property
        def raw_api(self) -> NCBlocksChildrenEndpoint:
            """Return the underlying endpoint in the Notion SDK."""
            return self.api.client.blocks.children

        # https://developers.notion.com/reference/patch-block-children
        def append(self, parent, *blocks: Block):
            """Add the given blocks as children of the specified parent.

            The blocks info will be updated based on returned data.

            `parent` may be any suitable `ObjectReference` type.
            """

            parent_id = ObjectReference.build(parent).id

            children = [block.serialize_for_api() for block in blocks if block is not None]

            logger.info('Appending %d blocks to %s ...', len(children), parent_id)

            # Typing within notion_client sucks, so we cast
            data = cast(dict[str, Any], self.raw_api.append(block_id=parent_id, children=children))

            if 'results' in data:
                if len(blocks) == len(data['results']):
                    for idx in range(len(blocks)):
                        block = blocks[idx]
                        result = data['results'][idx]
                        block.update(**result)

                else:
                    logger.warning('Unable to update results; size mismatch')

            else:
                logger.warning('Unable to update results; not provided')

            return parent

        # https://developers.notion.com/reference/get-block-children
        def list(self, parent):  # noqa: A003
            """Return all Blocks contained by the specified parent.

            `parent` may be any suitable `ObjectReference` type.
            """

            parent_id = ObjectReference.build(parent).id

            logger.info('Listing blocks for %s...', parent_id)

            blocks = EndpointIterator(endpoint=self.raw_api.list)

            return blocks(block_id=parent_id)

    def __init__(self, *args, **kwargs):
        """Initialize the `blocks` endpoint for the Notion API."""
        super().__init__(*args, **kwargs)

        self.children = BlocksEndpoint.ChildrenEndpoint(*args, **kwargs)

    @property
    def raw_api(self) -> NCBlocksEndpoint:
        """Return the underlying endpoint in the Notion SDK."""
        return self.api.client.blocks

    # https://developers.notion.com/reference/delete-a-block
    def delete(self, block: Block) -> Block:
        """Delete (archive) the specified Block.

        `block` may be any suitable `ObjectReference` type.
        """

        block_id = ObjectReference.build(block).id
        logger.info('Deleting block :: %s', block_id)

        data = self.raw_api.delete(block_id)

        return Block.model_validate(data)

    def restore(self, block: Block) -> Block:
        """Restore (unarchive) the specified Block.

        `block` may be any suitable `ObjectReference` type.
        """

        block_id = ObjectReference.build(block).id
        logger.info('Restoring block :: %s', block_id)

        data = self.raw_api.update(block_id, archived=False)

        return Block.model_validate(data)

    # https://developers.notion.com/reference/retrieve-a-block
    def retrieve(self, block) -> Block:
        """Return the requested Block.

        `block` may be any suitable `ObjectReference` type.
        """

        block_id = ObjectReference.build(block).id
        logger.info('Retrieving block :: %s', block_id)

        data = self.raw_api.retrieve(block_id)

        return Block.model_validate(data)

    # https://developers.notion.com/reference/update-a-block
    def update(self, block: Block) -> Block:
        """Update the block content on the server.

        The block info will be updated to the latest version from the server.
        """

        logger.info('Updating block :: %s', block.id)

        # Typing in notion_client sucks, so we cast
        data = cast(dict[str, Any], self.raw_api.update(block.id.hex, **block.serialize_for_api()))

        return block.update(**data)


class DatabasesEndpoint(Endpoint):
    """Interface to the 'databases' endpoint of the Notion API."""

    @property
    def raw_api(self) -> NCDatabasesEndpoint:
        """Return the underlying endpoint in the Notion SDK."""
        return self.api.client.databases

    @staticmethod
    def _build_request(
        parent: SerializeAsAny[ParentRef] | None = None,
        schema: dict[str, PropertyType] | None = None,
        title: list[RichTextObject] | None = None,
        description: list[RichTextObject] | None = None,
    ) -> dict[str, Any]:
        """Build a request payload from the given items.

        *NOTE* this method does not anticipate what the request will be used for and as
        such does not validate the inputs for any particular requests.
        """
        request = {}

        if parent is not None:
            request['parent'] = parent.serialize_for_api()

        if title is not None:
            request['title'] = [rt_obj.serialize_for_api() for rt_obj in title]

        if description is not None:
            request['description'] = [rt_obj.serialize_for_api() for rt_obj in description]

        if schema is not None:
            request['properties'] = {
                name: value.serialize_for_api() if value is not None else None for name, value in schema.items()
            }

        return request

    # https://developers.notion.com/reference/create-a-database
    def create(
        self, parent: Page, schema: dict[str, PropertyType], title: list[RichTextObject] | None = None
    ) -> Database:
        """Add a database to the given Page parent.

        `parent` may be any suitable `PageRef` type.
        """

        parent_ref = PageRef.build(parent)
        logger.info(f'Creating database `{title!s}` at `{parent_ref.page_id}`')

        request = self._build_request(parent_ref, schema, title)
        data = self.raw_api.create(**request)

        return Database.model_validate(data)

    # https://developers.notion.com/reference/retrieve-a-database
    def retrieve(self, dbref: Database | str | UUID) -> Database:
        """Return the Database with the given ID.

        `dbref` may be any suitable `DatabaseRef` type.
        """

        dbid = DatabaseRef.build(dbref).database_id
        logger.info(f'Retrieving database with id `{dbid}`')
        data = self.raw_api.retrieve(dbid)

        return Database.model_validate(data)

    def update(
        self,
        db: Database,
        title: list[RichTextObject] | None = None,
        description: list[RichTextObject] | None = None,
        schema: dict[str, PropertyType] | None = None,
    ) -> Database:
        """Update the Database object on the server.

        The database info will be updated to the latest version from the server.

        API reference: https://developers.notion.com/reference/update-a-database
        """
        logger.info(f'Updating database info of `{db.title}`')

        if request := self._build_request(schema=schema, title=title, description=description):
            # https://github.com/ramnes/notion-sdk-py/blob/main/notion_client/api_endpoints.py
            # Typing in notion_client sucks, thus we cast
            data = cast(dict[str, Any], self.raw_api.update(str(db.id), **request))
            db = db.update(**data)

        return db

    def delete(self, db: Database) -> Database:
        """Delete (archive) the specified Database."""

        dbid = DatabaseRef.build(db).database_id

        logger.info(f'Deleting database `{db}` with id {dbid}')

        block_obj = self.api.blocks.delete(dbid)
        # block.update(**data) is not possible as the API returns a block, not a database
        db.archived = block_obj.archived
        return db

    def restore(self, db: Database) -> Database:
        """Restore (unarchive) the specified Database."""

        dbid = DatabaseRef.build(db).database_id

        logger.info(f'Restoring database `{db}` with id {dbid}')

        block_obj = self.api.blocks.restore(dbid)
        # block.update(**data) is not possible as the API returns a block, not a database
        db.archived = block_obj.archived
        return db

    # https://developers.notion.com/reference/post-database-query
    def query(self, db: Database) -> DBQueryBuilder:
        """Initialize a new Query object with the target data class."""
        db_id = DatabaseRef.build(db).database_id
        logger.info('Initializing database query :: {%s}', db_id)

        return DBQueryBuilder(endpoint=self.raw_api.query, db_id=db_id)


class PagesEndpoint(Endpoint):
    """Interface to the API 'pages' endpoint."""

    class PropertiesEndpoint(Endpoint):
        """Interface to the API 'pages/properties' endpoint."""

        @property
        def raw_api(self):
            """Return the underlying endpoint in the Notion SDK"""
            return self.api.client.pages.properties

        # https://developers.notion.com/reference/retrieve-a-page-property
        def retrieve(self, page_id, property_id):
            """Return the Property on a specific Page with the given ID"""

            logger.info('Retrieving property :: %s [%s]', property_id, page_id)

            data = self.raw_api.retrieve(page_id, property_id)

            # TODO should PropertyListItem return an iterator instead?
            return TypeAdapter(PropertyItem | PropertyItemList).validate_python(data)

    def __init__(self, *args, **kwargs):
        """Initialize the `pages` endpoint for the Notion API"""
        super().__init__(*args, **kwargs)

        self.properties = PagesEndpoint.PropertiesEndpoint(*args, **kwargs)

    @property
    def raw_api(self):
        """Return the underlying endpoint in the Notion SDK"""
        return self.api.client.pages

    # https://developers.notion.com/reference/post-page
    def create(self, parent, title: Title | None = None, properties=None, children=None) -> Page:
        """Add a page to the given parent (Page or Database).

        `parent` may be a `ParentRef`, `Page`, or `Database` object.
        """

        if parent is None:
            msg = "'parent' must be provided"
            raise ValueError(msg)

        if isinstance(parent, Page):
            parent = PageRef.build(parent)
        elif isinstance(parent, Database):
            parent = DatabaseRef.build(parent)
        elif not isinstance(parent, ParentRef):
            msg = "Unsupported 'parent'"
            raise ValueError(msg)

        request = {'parent': parent.serialize_for_api()}

        # the API requires a properties object, even if empty
        if properties is None:
            properties = {}

        if title is not None:
            properties['title'] = title

        request['properties'] = {
            name: prop.serialize_for_api() if prop is not None else None for name, prop in properties.items()
        }

        if children is not None:
            request['children'] = [child.serialize_for_api() for child in children if child is not None]

        logger.info('Creating page :: %s => %s', parent, title)

        data = self.raw_api.create(**request)

        return Page.model_validate(data)

    def delete(self, page: Page) -> Page:
        """Delete (archive) the specified Page.

        `page` may be any suitable `PageRef` type.
        """

        return self.set_attr(page, archived=True)

    def restore(self, page: Page) -> Page:
        """Restore (unarchive) the specified Page.

        `page` may be any suitable `PageRef` type.
        """

        return self.set_attr(page, archived=False)

    # https://developers.notion.com/reference/retrieve-a-page
    def retrieve(self, page) -> Page:
        """Return the requested Page.

        `page` may be any suitable `PageRef` type.
        """

        page_id = PageRef.build(page).page_id

        logger.info('Retrieving page :: %s', page_id)

        data = self.raw_api.retrieve(page_id)

        # ToDo: would it make sense to (optionally) expand the full properties here?
        # e.g. call the PropertiesEndpoint to make sure all data is retrieved

        return Page.model_validate(data)

    # https://developers.notion.com/reference/patch-page
    def update(self, page: Page, **properties):
        """Update the Page object properties on the server.

        An optional `properties` may be specified as `"name"`: `PropertyValue` pairs.

        If `properties` are provided, only those values will be updated.
        If `properties` is empty, all page properties will be updated.

        The page info will be updated to the latest version from the server.
        """

        logger.info('Updating page info :: %s', page.id)

        if not properties:
            properties = page.properties

        props = {name: value.serialize_for_api() if value is not None else None for name, value in properties.items()}

        data = self.raw_api.update(page.id.hex, properties=props)

        return page.update(**data)

    def set_attr(
        self,
        page: Page,
        *,
        cover: FileObject | None | T_UNSET = UNSET,
        icon: FileObject | EmojiObject | None | T_UNSET = UNSET,
        archived: bool | T_UNSET = UNSET,
    ):
        """Set specific page attributes (such as cover, icon, etc.) on the server.

        `page` may be any suitable `PageRef` type.

        To remove an attribute, set its value to None.
        """

        page_id = PageRef.build(page).page_id

        props: dict[str, Any] = {}

        if cover is not UNSET:
            if cover is None:
                logger.info('Removing page cover :: %s', page_id)
                props['cover'] = None
            else:
                logger.info('Setting page cover :: %s => %s', page_id, cover)
                props['cover'] = cover.serialize_for_api()  # type: ignore[union-attr]

        if icon is not UNSET:
            if icon is None:
                logger.info('Removing page icon :: %s', page_id)
                props['icon'] = None
            else:
                logger.info('Setting page icon :: %s => %s', page_id, icon)
                props['icon'] = icon.serialize_for_api()  # type: ignore[union-attr]

        if archived is not UNSET:
            if archived:
                logger.info('Archiving page :: %s', page_id)
                props['archived'] = True
            else:
                logger.info('Restoring page :: %s', page_id)
                props['archived'] = False

        data = self.raw_api.update(page_id.hex, **props)

        return page.update(**data)


class SearchEndpoint(Endpoint):
    """Interface to the API 'search' endpoint."""

    # https://developers.notion.com/reference/post-search
    def __call__(self, text=None) -> SearchQueryBuilder:
        """Perform a search with the optional text"""
        return SearchQueryBuilder(endpoint=self.api.client.search, text=text)


class UsersEndpoint(Endpoint):
    """Interface to the API 'users' endpoint."""

    @property
    def raw_api(self):
        """Return the underlying endpoint in the Notion SDK"""
        return self.api.client.users

    # https://developers.notion.com/reference/get-users
    def as_list(self):
        """Return an iterator for all users in the workspace."""

        logger.info('Listing known users...')

        users = EndpointIterator(endpoint=self.raw_api.list)

        return users()

    # https://developers.notion.com/reference/get-user
    def retrieve(self, user_id):
        """Return the User with the given ID."""

        logger.info('Retrieving user :: %s', user_id)

        data = self.raw_api.retrieve(user_id)

        return User.model_validate(data)

    # https://developers.notion.com/reference/get-self
    def me(self):
        """Return the current bot User."""

        logger.info('Retrieving current integration bot')

        data = self.raw_api.me()

        return User.model_validate(data)



"""Provides an interactive query builder for Notion databases."""

from __future__ import annotations

import logging
from abc import ABC
from datetime import date, datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal
from uuid import UUID

from pydantic import ConfigDict, Field, SerializeAsAny, field_validator

from ultimate_notion.obj_api.core import GenericObject
from ultimate_notion.obj_api.iterator import MAX_PAGE_SIZE, EndpointIterator

if TYPE_CHECKING:
    from notion_client.api_endpoints import Endpoint as NCEndpoint

logger = logging.getLogger(__name__)


class TextCondition(GenericObject):
    """Represents text criteria in Notion."""

    equals: str | None = None
    does_not_equal: str | None = None
    contains: str | None = None
    does_not_contain: str | None = None
    starts_with: str | None = None
    ends_with: str | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class NumberCondition(GenericObject):
    """Represents number criteria in Notion."""

    equals: float | int | None = None
    does_not_equal: float | int | None = None
    greater_than: float | int | None = None
    less_than: float | int | None = None
    greater_than_or_equal_to: float | int | None = None
    less_than_or_equal_to: float | int | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class CheckboxCondition(GenericObject):
    """Represents checkbox criteria in Notion."""

    equals: bool | None = None
    does_not_equal: bool | None = None


class SelectCondition(GenericObject):
    """Represents select criteria in Notion."""

    equals: str | None = None
    does_not_equal: str | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class MultiSelectCondition(GenericObject):
    """Represents a multi_select criteria in Notion."""

    contains: str | None = None
    does_not_contains: str | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class DateCondition(GenericObject):
    """Represents date criteria in Notion."""

    equals: date | datetime | None = None
    before: date | datetime | None = None
    after: date | datetime | None = None
    on_or_before: date | datetime | None = None
    on_or_after: date | datetime | None = None

    is_empty: bool | None = None
    is_not_empty: bool | None = None

    past_week: Any | None = None
    past_month: Any | None = None
    past_year: Any | None = None
    next_week: Any | None = None
    next_month: Any | None = None
    next_year: Any | None = None


class PeopleCondition(GenericObject):
    """Represents people criteria in Notion."""

    contains: UUID | None = None
    does_not_contain: UUID | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class FilesCondition(GenericObject):
    """Represents files criteria in Notion."""

    is_empty: bool | None = None
    is_not_empty: bool | None = None


class RelationCondition(GenericObject):
    """Represents relation criteria in Notion."""

    contains: UUID | None = None
    does_not_contain: UUID | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class FormulaCondition(GenericObject):
    """Represents formula criteria in Notion."""

    string: TextCondition | None = None
    checkbox: CheckboxCondition | None = None
    number: NumberCondition | None = None
    date: DateCondition | None = None


class QueryFilter(GenericObject):
    """Base class for query filters."""


class PropertyFilter(QueryFilter):
    """Represents a database property filter in Notion."""

    property: str  # noqa: A003

    rich_text: TextCondition | None = None
    phone_number: TextCondition | None = None
    number: NumberCondition | None = None
    checkbox: CheckboxCondition | None = None
    select: SelectCondition | None = None
    multi_select: MultiSelectCondition | None = None
    date: DateCondition | None = None
    people: PeopleCondition | None = None
    files: FilesCondition | None = None
    relation: RelationCondition | None = None
    formula: FormulaCondition | None = None


class SearchFilter(QueryFilter):
    """Represents a search property filter in Notion."""

    property: str  # noqa: A003
    value: str


class TimestampKind(str, Enum):
    """Possible timestamp types."""

    CREATED_TIME = 'created_time'
    LAST_EDITED_TIME = 'last_edited_time'


class TimestampFilter(QueryFilter):
    """Represents a timestamp filter in Notion."""

    timestamp: TimestampKind


class CreatedTimeFilter(TimestampFilter):
    """Represents a created_time filter in Notion."""

    created_time: DateCondition
    timestamp: TimestampKind = TimestampKind.CREATED_TIME

    @classmethod
    def build(cls, value):
        """Create a new `CreatedTimeFilter` using the given constraint."""
        return CreatedTimeFilter(created_time=value)


class LastEditedTimeFilter(TimestampFilter):
    """Represents a last_edited_time filter in Notion."""

    last_edited_time: DateCondition
    timestamp: TimestampKind = TimestampKind.LAST_EDITED_TIME

    @classmethod
    def build(cls, value):
        """Create a new `LastEditedTimeFilter` using the given constraint."""
        return LastEditedTimeFilter(last_edited_time=value)


class CompoundFilter(QueryFilter):
    """Represents a compound filter in Notion."""

    # ToDo: Split this up in a And and Or Filter

    model_config = ConfigDict(populate_by_name=True)

    and_: list[QueryFilter] | None = Field(None, alias='and')
    or_: list[QueryFilter] | None = Field(None, alias='or')


class SortDirection(str, Enum):
    """Sort direction options."""

    ASCENDING = 'ascending'
    DESCENDING = 'descending'


class DBSort(GenericObject):
    """Sort instruction when querying a database"""

    property: str  # noqa: A003
    direction: SortDirection


class SearchSort(GenericObject):
    """Sort instruction when searching for pages and databases"""

    timestamp: Literal[TimestampKind.LAST_EDITED_TIME]
    direction: SortDirection


class Query(GenericObject, ABC):
    """Abstract query object in Notion for searching pages/databases and querying databases"""

    filter: SerializeAsAny[QueryFilter] | None = None  # noqa: A003
    start_cursor: UUID | None = None
    page_size: int = MAX_PAGE_SIZE

    @field_validator('page_size')
    @classmethod
    def valid_page_size(cls, value: int) -> int:
        """Validate that the given page size meets the Notion API requirements"""

        if value <= 0:
            msg = 'page size must be greater than zero'
            raise ValueError(msg)
        if value > MAX_PAGE_SIZE:
            msg = f'page size must be less than or equal to {MAX_PAGE_SIZE}'
            raise ValueError(msg)

        return value


class SearchQuery(Query):
    """Query object in Notion for searching pages & databases"""

    sort: SearchSort | None = None


class DBQuery(Query):
    """Query object in Notion for querying a database"""

    sorts: list[DBSort] | None = None


class QueryBuilder(ABC):
    """General query builder for the Notion search & database query API"""

    query: Query
    endpoint: NCEndpoint
    params: dict[str, str]

    def __init__(self, endpoint: NCEndpoint, **params: str | None):
        self.endpoint = endpoint
        self.params = {param: value for param, value in params.items() if value is not None}

    def start_at(self, cursor_id: UUID):
        """Set the start cursor to a specific page ID."""

        self.query.start_cursor = cursor_id
        return self

    def limit(self, count: int):
        """Limit the number of results to the given count."""

        self.query.page_size = count
        return self

    def execute(self):
        """Execute the current query and return an iterator for the results."""
        logger.debug('executing query - %s', self.query)

        # the API doesn't like "undefined" values...
        query = self.query.serialize_for_api()

        if self.params:
            query.update(self.params)

        return EndpointIterator(self.endpoint)(**query)

    def first(self):
        """Execute the current query and return the first result only."""

        try:
            return next(self.execute())
        except StopIteration:
            logger.debug('iterator returned empty result set')


class SearchQueryBuilder(QueryBuilder):
    """Search query builder to search for pages and databases

    By default and not changed by `sort`, then the most recently edited results are returned first.

    Notion API: https://developers.notion.com/reference/post-search
    """

    query: SearchQuery

    def __init__(self, endpoint, text: str | None = None):
        self.query = SearchQuery()
        super().__init__(endpoint=endpoint, query=text)

    def filter(self, *, page_only: bool = False, db_only: bool = False) -> SearchQueryBuilder:  # noqa: A003
        """Filter for pages or databases only"""
        if not (page_only ^ db_only):
            msg = 'Either `page_only` or `db_only` must be true, not both.'
            raise ValueError(msg)
        elif page_only:
            value = 'page'
        else:  # db_only
            value = 'database'

        self.query.filter = SearchFilter(property='object', value=value)
        return self

    def sort(self, *, ascending: bool) -> SearchQueryBuilder:
        """Add the given sort elements to the query."""
        direction = SortDirection.ASCENDING if ascending else SortDirection.DESCENDING
        self.query.sort = SearchSort(timestamp=TimestampKind.LAST_EDITED_TIME, direction=direction)
        return self


class DBQueryBuilder(QueryBuilder):
    """Query builder to query a database.

    Notion API: https://developers.notion.com/reference/post-search
    """

    query: DBQuery

    def __init__(self, endpoint, db_id: str):
        self.query = DBQuery()
        super().__init__(endpoint=endpoint, database_id=db_id)

    def filter(self, condition: QueryFilter) -> DBQueryBuilder:  # noqa: A003
        """Add the given filter to the query."""
        self.query.filter = condition
        return self

    def sort(self, sort_orders: DBSort | list[DBSort]) -> DBQueryBuilder:
        """Add the given sort elements to the query."""
        if isinstance(sort_orders, DBSort):
            sort_orders = [sort_orders]

        self.query.sorts = sort_orders
        return self



"""Functionality for working with Notion databases."""

from __future__ import annotations

from textwrap import dedent
from typing import cast

from ultimate_notion.blocks import DataObject
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.query import DBQueryBuilder
from ultimate_notion.objects import Emoji, File, RichText
from ultimate_notion.page import Page
from ultimate_notion.schema import Column, PageSchema, PropertyType, PropertyValue, ReadOnlyColumnError, SchemaError
from ultimate_notion.text import camel_case, snake_case
from ultimate_notion.utils import dict_diff_str, get_active_session, get_repr
from ultimate_notion.view import View


class Database(DataObject[obj_blocks.Database], wraps=obj_blocks.Database):
    """A Notion database.

    This object always represents an original database, not a linked database.

    API reference: https://developers.notion.com/docs/working-with-databases
    """

    _schema: type[PageSchema] | None = None

    def __str__(self):
        if self.title:
            return str(self.title)
        else:
            return 'Untitled'

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically"""
        return str(self)

    def __repr__(self) -> str:
        return get_repr(self)

    @property
    def title(self) -> RichText:
        """Return the title of this database as rich text."""
        title = self.obj_ref.title
        return RichText.wrap_obj_ref(title)

    @title.setter
    def title(self, text: str | RichText):
        """Set the title of this database"""
        if not isinstance(text, RichText):
            text = RichText.from_plain_text(text)
        session = get_active_session()
        session.api.databases.update(self.obj_ref, title=text.obj_ref)

    @property
    def description(self) -> RichText:
        """Return the description of this database as rich text."""
        desc = self.obj_ref.description
        return RichText.wrap_obj_ref(desc)

    @description.setter
    def description(self, text: str | RichText):
        """Set the description of this database."""
        if not isinstance(text, RichText):
            text = RichText.from_plain_text(text)
        session = get_active_session()
        session.api.databases.update(self.obj_ref, description=text.obj_ref)

    @property
    def icon(self) -> File | Emoji | None:
        icon = self.obj_ref.icon
        if isinstance(icon, objs.FileObject):
            return File.wrap_obj_ref(icon)
        elif isinstance(icon, objs.EmojiObject):
            return Emoji.wrap_obj_ref(icon)
        elif icon is None:
            return None
        else:
            msg = f'unknown icon object of {type(icon)}'
            raise RuntimeError(msg)

    @property
    def cover(self) -> File | None:
        cover = self.obj_ref.cover
        return File.wrap_obj_ref(cover) if cover is not None else None

    @property
    def is_wiki(self) -> bool:
        """Is this database a wiki database."""
        # ToDo: Implement using the verification property
        raise NotImplementedError

    def _reflect_schema(self, obj_ref: obj_blocks.Database) -> type[PageSchema]:
        """Reflection about the database schema."""
        title = str(self)
        cls_name = f'{camel_case(title)}Schema'
        attrs = {
            snake_case(k): Column(k, cast(PropertyType, PropertyType.wrap_obj_ref(v)))
            for k, v in obj_ref.properties.items()
        }
        schema: type[PageSchema] = type(cls_name, (PageSchema,), attrs, db_title=title)
        schema.bind_db(self)
        return schema

    @property
    def schema(self) -> type[PageSchema]:
        """Schema of the database."""
        if not self._schema:
            self._schema = self._reflect_schema(self.obj_ref)
        return self._schema

    @schema.setter
    def schema(self, schema: type[PageSchema]):
        """Set a custom schema in order to change the Python variables names."""
        if self.schema.is_consistent_with(schema):
            schema.bind_db(self)
            self._schema = schema
        else:
            cols_added, cols_removed, cols_changed = dict_diff_str(self.schema.to_dict(), schema.to_dict())
            msg = f"""Provided schema is not consistent with the current schema of the database:
                      Columns added: {cols_added}
                      Columns removed: {cols_removed}
                      Columns changed: {cols_changed}
                   """
            raise SchemaError(dedent(msg))

    @property
    def url(self) -> str:
        return self.obj_ref.url

    @property
    def is_inline(self) -> bool:
        return self.obj_ref.is_inline

    def delete(self) -> Database:
        """Delete/archive this database."""
        if not self.is_deleted:
            session = get_active_session()
            session.api.databases.delete(self.obj_ref)
        return self

    def restore(self) -> Database:
        """Restore/unarchive this database."""
        if self.is_deleted:
            session = get_active_session()
            session.api.databases.restore(self.obj_ref)
        return self

    def reload(self) -> Database:
        """Reload this database."""
        session = get_active_session()
        self.obj_ref = session.api.databases.retrieve(self.obj_ref.id)
        self.schema._set_obj_refs()
        return self

    @staticmethod
    def _pages_from_query(query: DBQueryBuilder) -> list[Page]:
        # ToDo: Remove when self.query is implemented!
        cache = get_active_session().cache
        pages = []
        for page_obj in query.execute():
            if page_obj.id in cache:
                page = cast(Page, cache[page_obj.id])
                page.obj_ref = page_obj  # updates the page content
            else:
                page = Page.wrap_obj_ref(page_obj)
            pages.append(page)
        return pages

    def fetch_all(self) -> View:
        """Fetch all pages and return a view."""
        session = get_active_session()
        query = session.api.databases.query(self.obj_ref)  # ToDo: use self.query when implemented
        return View(database=self, pages=self._pages_from_query(query), query=query)

    def create_page(self, **kwargs) -> Page:
        """Create a page with properties according to the schema within the corresponding database."""
        schema_kwargs = {col.attr_name: col for col in self.schema.get_cols()}
        if not set(kwargs).issubset(set(schema_kwargs)):
            add_kwargs = set(kwargs) - set(schema_kwargs)
            msg = f"kwargs {', '.join(add_kwargs)} not defined in schema"
            raise SchemaError(msg)

        schema_dct = {}
        for kwarg, value in kwargs.items():
            col = schema_kwargs[kwarg]
            prop_value_cls = col.type.prop_value  # map schema to page property
            # ToDo: Check at that point in case of selectoption if the option is already defined in Schema!

            if prop_value_cls.readonly:
                raise ReadOnlyColumnError(col)

            prop_value = value if isinstance(value, PropertyValue) else prop_value_cls(value)

            schema_dct[schema_kwargs[kwarg].name] = prop_value.obj_ref

        session = get_active_session()
        page = Page.wrap_obj_ref(session.api.pages.create(parent=self.obj_ref, properties=schema_dct))
        return page

    def query(self):
        """Query a (large) database for pages in a more specific way."""
        raise NotImplementedError



"""Ultimate Notion provides a pythonic, high-level API for Notion.

Notion-API: https://developers.notion.com/reference/intro
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version('ultimate-notion')
except PackageNotFoundError:  # pragma: no cover
    __version__ = 'unknown'
finally:
    del version, PackageNotFoundError

from ultimate_notion.database import Database
from ultimate_notion.obj_api.enums import AggFunc, BGColor, CodeLang, Color, NumberFormat, VState
from ultimate_notion.objects import Emoji, File, Option, OptionNS, RichText, User
from ultimate_notion.page import Page
from ultimate_notion.schema import ColType, Column, PageSchema, SelfRef
from ultimate_notion.session import Session

__all__ = [
    'AggFunc',
    'BGColor',
    'CodeLang',
    'Color',
    'ColType',
    'Column',
    'Database',
    'Emoji',
    'File',
    'NumberFormat',
    'Option',
    'OptionNS',
    'Page',
    'PageSchema',
    'RichText',
    'SelfRef',
    'Session',
    'User',
    'VState',
    '__version__',
]



"""Functionality for general Notion objects like texts, files, options, etc."""

from __future__ import annotations

from typing import cast

from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.enums import Color
from ultimate_notion.text import chunky, html_img
from ultimate_notion.utils import Wrapper, get_repr


class Option(Wrapper[objs.SelectOption], wraps=objs.SelectOption):
    """Option for select & multi-select property."""

    def __init__(self, name: str, *, color: Color = Color.DEFAULT):
        super().__init__(name, color=color)

    @property
    def id(self) -> str:  # noqa: A003
        """ID of the option."""
        return self.obj_ref.id

    @property
    def name(self) -> str:
        """Name of the option."""
        return self.obj_ref.name

    def description(self) -> str:
        """Description of the option."""
        if desc := self.obj_ref.description:
            return desc
        else:
            return ''

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.name


class OptionNS:
    """Option namespace to simplify working with (Multi-)Select options."""

    @classmethod
    def to_list(cls) -> list[Option]:
        """Convert the enum to a list as needed by the (Multi)Select column types."""
        return [
            getattr(cls, var) for var in cls.__dict__ if not var.startswith('__') and not callable(getattr(cls, var))
        ]


class OptionGroup(Wrapper[objs.SelectGroup], wraps=objs.SelectGroup):
    """Group of options for status property."""

    _options: dict[str, Option]  # holds all possible options

    @classmethod
    def wrap_obj_ref(cls, obj_ref, /, *, options: list[Option] | None = None) -> OptionGroup:
        """Convienence constructor for the group of options."""
        obj = super().wrap_obj_ref(obj_ref)
        options = [] if options is None else options
        obj._options = {option.id: option for option in options}
        return obj

    @property
    def name(self) -> str:
        """Name of the option group."""
        return self.obj_ref.name

    @property
    def options(self) -> list[Option]:
        """Options within this option group."""
        return [self._options[opt_id] for opt_id in self.obj_ref.option_ids]

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.name


class File(Wrapper[objs.FileObject], wraps=objs.FileObject):
    """A web resource e.g. for the files property."""

    obj_ref: objs.FileObject

    @classmethod
    def wrap_obj_ref(cls, obj_ref: objs.FileObject) -> File:
        self = cast(File, cls.__new__(cls))
        self.obj_ref = obj_ref
        return self

    def __init__(self, *, url: str, name: str | None = None) -> None:
        self.obj_ref = objs.ExternalFile.build(url=url, name=name)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return str(self) == other
        elif isinstance(other, File):
            return str(self) == str(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(str(self))

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.url

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically to display this file."""
        return html_img(self.url, size=2)

    @property
    def name(self) -> str | None:
        return self.obj_ref.name

    @property
    def url(self) -> str:
        if isinstance(self.obj_ref, objs.HostedFile):
            return self.obj_ref.file.url
        elif isinstance(self.obj_ref, objs.ExternalFile):
            return self.obj_ref.external.url
        else:
            msg = f'Unknown file type {type(self.obj_ref)}'
            raise RuntimeError(msg)


class Emoji(Wrapper[objs.EmojiObject], str, wraps=objs.EmojiObject):
    """Emoji object which behaves like str."""

    def __init__(self, emoji: str) -> None:
        self.obj_ref = objs.EmojiObject.build(emoji)

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.obj_ref.emoji

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return str(self) == other
        elif isinstance(other, Emoji):
            return str(self) == str(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(str(self))

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically to display this file."""
        return str(self)

    def to_code(self) -> str:
        """Represent the emoji as :shortcode:, e.g. :smile:"""
        raise NotImplementedError

    @classmethod
    def from_code(cls, shortcode: str) -> Emoji:
        """Create an Emoji object from a :shortcode:, e.g. :smile:"""
        raise NotImplementedError


class RichTextBase(Wrapper[objs.RichTextObject], wraps=objs.RichTextObject):
    """Super class for text, equation and mentions of various kinds."""


class Text(RichTextBase, wraps=objs.TextObject):
    """A Text object."""


class Equation(RichTextBase, wraps=objs.EquationObject):
    """An Equation object."""


class Mention(RichTextBase, wraps=objs.MentionObject):
    """A Mention object."""


class RichText(str):
    """User-facing class holding several RichTexts."""

    rich_texts: list[RichTextBase]

    def __init__(self, plain_text: str):
        super().__init__()

        rich_texts: list[RichTextBase] = []
        for part in chunky(plain_text):
            rich_texts.append(Text(part))
        self.rich_texts = rich_texts

    @classmethod
    def wrap_obj_ref(cls, obj_refs: list[objs.RichTextObject]) -> RichText:
        rich_texts = [cast(RichTextBase, RichTextBase.wrap_obj_ref(obj_ref)) for obj_ref in obj_refs]
        plain_text = ''.join(text.obj_ref.plain_text for text in rich_texts if text)
        obj = cls(plain_text)
        obj.rich_texts = rich_texts
        return obj

    @property
    def obj_ref(self) -> list[objs.RichTextObject]:
        return [elem.obj_ref for elem in self.rich_texts]

    @classmethod
    def from_markdown(cls, text: str) -> RichText:
        """Create RichTextList by parsing the markdown."""
        # ToDo: Implement
        # ToDo: Handle Equations and Mentions here accordingly
        raise NotImplementedError

    def to_markdown(self) -> str | None:
        """Convert the list of RichText objects to markdown."""
        # ToDo: Implement
        raise NotImplementedError

    @classmethod
    def from_plain_text(cls, text: str) -> RichText:
        """Create RichTextList from plain text.

        This method is a more explicit alias for the default constructor
        """
        return cls(text)

    def to_plain_text(self) -> str:
        """Return rich text as plaintext

        This method is a more explicit variant then just using the object.
        """
        return str(self)

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically to display this text."""
        # ToDo: Later use Markdown output.
        return self.to_plain_text()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return str(self) == other
        elif isinstance(other, RichText):
            return str(self) == str(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(str(self))


class User(Wrapper[objs.User], wraps=objs.User):
    @classmethod
    def wrap_obj_ref(cls, obj_ref: objs.User) -> User:
        self = cast(User, cls.__new__(cls))
        self.obj_ref = obj_ref
        return self

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return get_repr(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, User):
            return NotImplemented
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def id(self):  # noqa: A003
        return self.obj_ref.id

    @property
    def name(self):
        return self.obj_ref.name

    @property
    def is_person(self) -> bool:
        return isinstance(self.obj_ref, objs.Person)

    @property
    def is_bot(self) -> bool:
        return isinstance(self.obj_ref, objs.Bot)

    @property
    def avatar_url(self):
        return self.obj_ref.avatar_url

    @property
    def email(self) -> str | None:
        if isinstance(self.obj_ref, objs.Person):
            return self.obj_ref.person.email
        else:  # it's a bot without an e-mail
            return None



"""
Page properties especially for pages within databases.
The names of the properties reflect the name in the Notion UI.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

import ultimate_notion.obj_api.props as obj_props
from ultimate_notion.obj_api.props import DateType
from ultimate_notion.objects import File, Option, RichText, User
from ultimate_notion.utils import Wrapper, get_active_session, get_repr

if TYPE_CHECKING:
    from ultimate_notion.page import Page

T = TypeVar('T', bound=obj_props.PropertyValue)


class PropertyValue(Wrapper[T], ABC, wraps=obj_props.PropertyValue):  # noqa: PLW1641
    """Base class for Notion property values.

    Used to map high-level objects to low-level Notion-API objects
    """

    readonly: bool = False  # value of property can not be set by us
    _type_value_map: ClassVar[dict[str, type[PropertyValue]]] = {}

    def __init_subclass__(cls, wraps: type[T], **kwargs: Any):
        super().__init_subclass__(wraps=wraps, **kwargs)
        # at that time the model is not yet constructed, thus no direct field acces with .type.
        type_name = wraps.model_fields['type'].get_default()
        cls._type_value_map[type_name] = cls

    @property
    def _obj_api_type(self) -> type[obj_props.PropertyValue]:
        return self._obj_api_map_inv[self.__class__]

    def __init__(self, values):
        if isinstance(values, list):
            values = [value.obj_ref if isinstance(value, Wrapper) else value for value in values]
        else:
            values = values.obj_ref if isinstance(values, Wrapper) else values

        super().__init__(values)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PropertyValue):
            return NotImplemented
        return self.obj_ref.type == other.obj_ref.type and self.obj_ref.value == self.obj_ref.value

    @property
    @abstractmethod
    def value(self) -> Any:
        """Return the actual Python value object of this property."""

    @property
    def id(self) -> str:  # noqa: A003
        return self.obj_ref.id

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        if isinstance(self.value, RichText):
            return str(self.value) if self.value else ''
        elif isinstance(self.value, list):
            # workaround as `str` on lists calls `repr` instead of `str`
            return ', '.join(str(val) for val in self.value)
        else:
            return str(self.value) if self.value else ''


class Title(PropertyValue[obj_props.Title], wraps=obj_props.Title):
    """Title property value."""

    def __init__(self, text: str | RichText):
        if isinstance(text, str):
            text = RichText.from_plain_text(text)
        super().__init__(text.obj_ref)

    @property
    def value(self) -> RichText:
        return RichText.wrap_obj_ref(self.obj_ref.title)


class Text(PropertyValue[obj_props.RichText], wraps=obj_props.RichText):
    """Rich text property value."""

    def __init__(self, text: str | RichText):
        if isinstance(text, str):
            text = RichText.from_plain_text(text)
        super().__init__(text.obj_ref)

    @property
    def value(self) -> RichText:
        return RichText.wrap_obj_ref(self.obj_ref.rich_text)


class Number(PropertyValue[obj_props.Number], wraps=obj_props.Number):
    """Number property value."""

    @property
    def value(self) -> int | float | None:
        return self.obj_ref.number


class Checkbox(PropertyValue[obj_props.Checkbox], wraps=obj_props.Checkbox):
    """Simple checkbox type; represented as a boolean."""

    @property
    def value(self) -> bool | None:
        return self.obj_ref.checkbox


class Date(PropertyValue[obj_props.Date], wraps=obj_props.Date):
    """Date(-time) property value."""

    # ToDo: Also use `time_zone` here defined in obj_api props or objects
    def __init__(self, start: datetime | date, end: datetime | date | None = None):
        self.obj_ref = obj_props.Date.build(start, end)

    @property
    def value(self) -> DateType | None:
        # ToDo: Set `time_zone` accordingly if provided
        date = self.obj_ref.date
        if date is None:
            return None
        elif date.end is None:
            return date.start
        else:
            return date.start, date.end


class Status(PropertyValue[obj_props.Status], wraps=obj_props.Status):
    """Status property value."""

    def __init__(self, option: str | Option):
        if isinstance(option, str):
            option = Option(option)

        super().__init__(option)

    @property
    def value(self) -> Option | None:
        if self.obj_ref.status:
            return Option.wrap_obj_ref(self.obj_ref.status)
        else:
            return None


class Select(PropertyValue[obj_props.Select], wraps=obj_props.Select):
    """Single select property value."""

    def __init__(self, option: str | Option):
        if isinstance(option, str):
            option = Option(option)

        super().__init__(option)

    @property
    def value(self) -> Option | None:
        if self.obj_ref.select:
            return Option.wrap_obj_ref(self.obj_ref.select)
        else:
            return None


class MultiSelect(PropertyValue[obj_props.MultiSelect], wraps=obj_props.MultiSelect):
    """Multi-select property value."""

    def __init__(self, options: str | Option | list[str | Option]):
        if not isinstance(options, list):
            options = [options]
        options = [Option(option) if isinstance(option, str) else option for option in options]
        super().__init__(options)

    @property
    def value(self) -> list[Option] | None:
        if self.obj_ref.multi_select:
            return [Option.wrap_obj_ref(option) for option in self.obj_ref.multi_select]
        else:
            return None


class People(PropertyValue[obj_props.People], wraps=obj_props.People):
    """People property value."""

    def __init__(self, users: User | list[User]):
        if not isinstance(users, list):
            users = [users]
        super().__init__(users)

    @property
    def value(self) -> list[User]:
        session = get_active_session()
        return [session.get_user(user.id) for user in self.obj_ref.people]


class URL(PropertyValue[obj_props.URL], wraps=obj_props.URL):
    """URL property value."""

    @property
    def value(self) -> str | None:
        return self.obj_ref.url


class Email(PropertyValue[obj_props.Email], wraps=obj_props.Email):
    """Email property value."""

    @property
    def value(self) -> str | None:
        return self.obj_ref.email


class PhoneNumber(PropertyValue[obj_props.PhoneNumber], wraps=obj_props.PhoneNumber):
    """Phone property value."""

    @property
    def value(self) -> str | None:
        return self.obj_ref.phone_number


class Files(PropertyValue[obj_props.Files], wraps=obj_props.Files):
    """Files property value."""

    def __init__(self, files: File | list[File]):
        if not isinstance(files, list):
            files = [files]

        super().__init__(files)

    @property
    def value(self) -> list[File]:
        return [File.wrap_obj_ref(file) for file in self.obj_ref.files]


class Formula(PropertyValue[obj_props.Formula], wraps=obj_props.Formula):
    """Formula property value."""

    readonly = True

    @property
    def value(self) -> str | float | int | DateType | None:  # all values of subclasses of `FormulaResult`
        return self.obj_ref.formula.value if self.obj_ref.formula else None


class Relations(PropertyValue[obj_props.Relation], wraps=obj_props.Relation):
    """Relation property value."""

    def __init__(self, pages: Page | list[Page]):
        if not isinstance(pages, list):
            pages = [pages]
        super().__init__(pages)

    @property
    def value(self) -> list[Page]:
        session = get_active_session()
        return [session.get_page(ref_obj.id) for ref_obj in self.obj_ref.relation]


class Rollup(PropertyValue[obj_props.Rollup], wraps=obj_props.Rollup):
    """Rollup property value."""

    readonly = True

    @property
    def value(self) -> str | float | int | DateType | list[PropertyValue] | None:
        # ToDo: Write a unit test for this!
        if self.obj_ref.rollup is None:
            return None

        rollup_val = self.obj_ref.rollup.value
        if isinstance(rollup_val, obj_props.RollupArray):
            return [PropertyValue.wrap_obj_ref(prop) for prop in rollup_val]
        else:
            return rollup_val


class CreatedTime(PropertyValue[obj_props.CreatedTime], wraps=obj_props.CreatedTime):
    """Created-time property value."""

    readonly = True


class CreatedBy(PropertyValue[obj_props.CreatedBy], wraps=obj_props.CreatedBy):
    """Created-by property value."""

    readonly = True


class LastEditedTime(PropertyValue[obj_props.LastEditedTime], wraps=obj_props.LastEditedTime):
    """Last-edited-time property value."""

    readonly = True


class LastEditedBy(PropertyValue[obj_props.LastEditedBy], wraps=obj_props.LastEditedBy):
    """Last-edited-by property value."""

    readonly = True


class ID(PropertyValue[obj_props.UniqueID], wraps=obj_props.UniqueID):
    """Unique ID property value."""

    readonly = True


class Verification(PropertyValue[obj_props.Verification], wraps=obj_props.Verification):
    """Verification property value of pages in wiki databases."""

    readonly = True
    # ToDo: Implement properties to retrieve user, etc. and convert user to actual high-level User object



"""Functionality around Notion pages."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

import mistune
from emoji import is_emoji

from ultimate_notion.blocks import Block, ChildDatabase, ChildPage, DataObject
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api import props as obj_props
from ultimate_notion.objects import Emoji, File, RichText
from ultimate_notion.props import PropertyValue, Title
from ultimate_notion.utils import get_active_session, get_repr, is_notebook

if TYPE_CHECKING:
    from ultimate_notion.database import Database


class PageProperty:
    """Property of a page implementing the descriptor protocol."""

    def __init__(self, prop_name: str):
        self._prop_name = prop_name

    def __get__(self, obj: PageProperties, type=None) -> PropertyValue:  # noqa: A002
        return obj[self._prop_name]

    def __set__(self, obj: PageProperties, value):
        obj[self._prop_name] = value


class PageProperties:
    """Properties of a page as defined in the schema of the database.

    This defines the `.props` namespace of a page `page` and updates the content
    on the Notion server side in case of an assignment.
    Access the properties with `page.props.property_name` or `page.props["Property Name"]`.
    """

    def __init__(self, page: Page):
        self._page = page

    @property
    def _properties(self) -> dict[str, obj_props.PropertyValue]:
        """Return the low-level page properties"""
        return self._page.obj_ref.properties

    def __getitem__(self, prop_name: str) -> PropertyValue:
        prop = self._properties.get(prop_name)
        if prop is None:
            msg = f'No such property: {prop_name}'
            raise AttributeError(msg)

        return cast(PropertyValue, PropertyValue.wrap_obj_ref(prop))

    def __setitem__(self, prop_name: str, value: Any):
        # Todo: use the schema of the database to see which properties are writeable at all.
        db = self._page.database
        if db is None:
            msg = f'Trying to set a property but page {self._page} is not bound to any database'
            raise RuntimeError(msg)

        if not isinstance(value, PropertyValue):
            # construct concrete PropertyValue using the schema
            prop_type = db.schema.to_dict()[prop_name]
            value = prop_type.prop_value(value)

        # update the property on the server (which will update the local data)
        session = get_active_session()
        session.api.pages.update(self._page.obj_ref, **{prop_name: value.obj_ref})

    def __iter__(self):
        """Iterator of property names."""
        yield from self._properties.keys()

    def to_dict(self) -> dict[str, PropertyValue]:
        """All page properties as dictionary."""
        return {prop_name: self[prop_name] for prop_name in self}


class Page(DataObject[obj_blocks.Page], wraps=obj_blocks.Page):
    """A Notion page.

    Attributes:
        props: accessor for all page properties
    """

    props: PageProperties
    _render_md = mistune.create_markdown(plugins=['strikethrough', 'url', 'task_lists', 'math'], escape=False)
    _content: list[Block] | None = None

    @classmethod
    def wrap_obj_ref(cls, obj_ref: obj_blocks.Page, /) -> Page:
        obj = super().wrap_obj_ref(obj_ref)
        obj.props = obj._create_prop_attrs()
        return obj

    def _create_prop_attrs(self) -> PageProperties:
        """Create the attributes for the database properties of this page."""
        # We have to subclass in order to populate it with the descriptor `PageProperty``
        # as this only works on the class level and we want a unique class for each property.
        page_props_cls = type('_PageProperties', (PageProperties,), {})
        if self.database:
            for col in self.database.schema.get_cols():
                setattr(page_props_cls, col.attr_name, PageProperty(prop_name=col.name))
        return page_props_cls(page=self)

    def __str__(self) -> str:
        return str(self.title)

    def __repr__(self) -> str:
        return get_repr(self, desc=self.title)

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically to display this page."""
        return self._render_md(self.to_markdown())  # Github flavored markdown

    @property
    def content(self) -> list[Block]:
        """Return the content of this page, i.e. all blocks belonging to this page"""
        if self._content is None:  # generate cache
            session = get_active_session()
            self._content = session._get_blocks(self.id)

        return self._content

    @property
    def children(self) -> list[Page | Database]:
        """Return all contained databases and pages within this page"""
        session = get_active_session()
        children: list[Page | Database] = []
        for block in self.content:
            if isinstance(block, ChildPage):
                children.append(session.get_page(block.id))
            elif isinstance(block, ChildDatabase):
                children.append(session.get_db(block.id))
        return children

    @property
    def database(self) -> Database | None:
        """If this page is located in a database return the database or None otherwise."""
        from ultimate_notion.database import Database  # noqa: PLC0415

        if isinstance(self.parent, Database):
            return self.parent
        else:
            return None

    def _get_title_col_name(self) -> str:
        """Get the name of the title property."""
        # As the 'title' property might be renamed in case of pages in databases, we look for the `id`.
        for name, prop in self.obj_ref.properties.items():
            if prop.id == 'title':
                return name
        msg = 'Encountered a page without title property'
        raise RuntimeError(msg)

    @property
    def title(self) -> RichText:
        """Title of the page."""
        title_prop_name = self._get_title_col_name()
        title_prop = self.obj_ref.properties[title_prop_name]
        title = cast(Title, PropertyValue.wrap_obj_ref(title_prop))
        return title.value

    @title.setter
    def title(self, text: RichText | str | None):
        """Set the title of the page."""
        if text is None:
            text = ''
        title = Title(text)
        title_prop_name = self._get_title_col_name()
        session = get_active_session()
        session.api.pages.update(self.obj_ref, **{title_prop_name: title.obj_ref})

    @property
    def icon(self) -> File | Emoji | None:
        """Icon of the page, i.e. emojis, Notion's icons, or custom images."""
        icon = self.obj_ref.icon
        if isinstance(icon, objs.ExternalFile):
            return File.wrap_obj_ref(icon)
        elif isinstance(icon, objs.EmojiObject):
            return Emoji.wrap_obj_ref(icon)
        elif icon is None:
            return None
        else:
            msg = f'unknown icon object of {type(icon)}'
            raise RuntimeError(msg)

    @icon.setter
    def icon(self, icon: File | Emoji | str | None):
        """Set the icon of this page."""
        if isinstance(icon, str) and not isinstance(icon, Emoji):
            icon = Emoji(icon) if is_emoji(icon) else File(url=icon, name=None)
        icon_obj = None if icon is None else icon.obj_ref
        session = get_active_session()
        session.api.pages.set_attr(self.obj_ref, icon=icon_obj)

    @property
    def cover(self) -> File | None:
        """Cover of the page."""
        cover = self.obj_ref.cover
        if isinstance(cover, objs.ExternalFile):
            return File.wrap_obj_ref(cover)
        elif cover is None:
            return None
        else:
            msg = f'unknown cover object of {type(cover)}'
            raise RuntimeError(msg)

    @cover.setter
    def cover(self, cover: File | str | None):
        """Set the cover fo this page."""
        if isinstance(cover, str):
            cover = File(url=cover, name=None)
        cover_obj = None if cover is None else cover.obj_ref
        session = get_active_session()
        session.api.pages.set_attr(self.obj_ref, cover=cover_obj)

    def to_markdown(self) -> str:
        """Return the content of the page as Markdown."""
        # ToDo: Since notion2md is also quite buggy, use an own implementation here using Mistune!
        import os  # noqa: PLC0415

        from notion2md.exporter.block import StringExporter  # noqa: PLC0415

        from ultimate_notion.session import ENV_NOTION_TOKEN, Session  # noqa: PLC0415

        auth_token = Session.get_or_create().client.options.auth
        if auth_token is None:
            msg = 'Impossible to pass the auth token to notion_client via an environment variable'
            raise RuntimeError(msg)
        os.environ[ENV_NOTION_TOKEN] = auth_token  # used internally by notion_client used by notion2md

        md = f'# {self.title}\n'
        md += StringExporter(block_id=self.id).export()
        return md

    def show(self, *, simple: bool | None = None):
        """Show the content of the page, rendered in Jupyter Lab"""
        simple = simple if simple is not None else is_notebook()
        md = self.to_markdown()

        if simple:
            print(md)  # noqa: T201
        else:
            from IPython.core.display import display_markdown  # noqa: PLC0415

            display_markdown(md, raw=True)

    def delete(self) -> Page:
        """Delete/archive this page."""
        if not self.is_deleted:
            session = get_active_session()
            session.api.pages.delete(self.obj_ref)
        return self

    def restore(self) -> Page:
        """Restore/unarchive this page."""
        if self.is_deleted:
            session = get_active_session()
            session.api.pages.restore(self.obj_ref)
        return self

    def reload(self) -> Page:
        """Reload this page."""
        session = get_active_session()
        self.obj_ref = session.api.pages.retrieve(self.obj_ref.id)
        self._content = None  # this forces a new retrieval next time
        return self



"""The Notion API mapped to Python objects by Pydantic.

The code was taken originally from [Notional] by Jason Heddings and is MIT-licensed.
Due to the Pydantic v2 migration and several other design changes, it was refactored since then.

[Notional]: https://github.com/jheddings/notional
"""



"""Additional utilities that fit nowhere else."""

from __future__ import annotations

import textwrap
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeAlias, TypeVar, cast
from uuid import UUID

import numpy as np

from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.core import GenericObject

if TYPE_CHECKING:
    from ultimate_notion.session import Session


ObjRef: TypeAlias = UUID | str


T = TypeVar('T')


class SList(list[T]):
    """A list that holds often only a single element."""

    def item(self) -> T:
        if len(self) == 1:
            return self[0]
        elif len(self) == 0:
            msg = 'list is empty'
        else:
            msg = f"list of '{type(self[0]).__name__}' objects has more than one element"
        raise ValueError(msg)


def is_notebook() -> bool:
    """Determine if we are running within a Jupyter notebook."""
    try:
        from IPython import get_ipython  # noqa: PLC0415
    except ModuleNotFoundError:
        return False  # Probably standard Python interpreter
    else:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)


def store_retvals(func):
    """Decorator storing the return values as function attribute for later cleanups.

    This can be used for instance in a generator like this:
    ```
    @pytest.fixture
    def create_blank_db(notion, test_area):
        @store_retvals
        def nested_func(db_name):
            db = notion.databases.create(
                parent=test_area,
                title=db_name,
                schema={
                    "Name": schema.Title(),
                },
            )
            return db

        yield nested_func

        # clean up by deleting the db of each prior call
        for db in nested_func.retvals:
            notion.databases.delete(db)
    ```
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        retval = func(*args, **kwargs)
        wrapped.retvals.append(retval)
        return retval

    wrapped.retvals = []
    return wrapped


def find_indices(elements: np.ndarray | list[Any], total_set: np.ndarray | list[Any]) -> np.ndarray:
    """Finds the indices of the elements in the total set."""
    if not isinstance(total_set, np.ndarray):
        total_set = np.array(total_set)
    mask = np.isin(total_set, elements)
    indices = np.where(mask)[0]
    lookup = dict(zip(total_set[mask], indices, strict=True))
    result = np.array([lookup.get(x) for x in elements])
    return result


def find_index(elem: Any, lst: list[Any]) -> int | None:
    """Find the index of the element in the list or return `None`."""
    if elem not in lst:
        return None
    else:
        return lst.index(elem)


def deepcopy_with_sharing(obj: Any, shared_attributes: list[str], memo: dict[int, Any] | None = None):
    """Like `deepcopy` but specified attributes are shared.

    Deepcopy an object, except for a given list of attributes, which should
    be shared between the original object and its copy.

    Args:
        obj: some object to copy
        shared_attributes: A list of strings identifying the attributes that should be shared instead of copied.
        memo: dictionary passed into __deepcopy__.  Ignore this argument if not calling from within __deepcopy__.

    Example:
        ```python
        class A(object):
            def __init__(self):
                self.copy_me = []
                self.share_me = []

            def __deepcopy__(self, memo):
                return deepcopy_with_sharing(
                    self, shared_attribute_names=["share_me"], memo=memo
                )


        a = A()
        b = deepcopy(a)
        assert a.copy_me is not b.copy_me
        assert a.share_me is b.share_me

        c = deepcopy(b)
        assert c.copy_me is not b.copy_me
        assert c.share_me is b.share_me
        ```

    Original from https://stackoverflow.com/a/24621200
    """
    shared_attrs = {k: getattr(obj, k) for k in shared_attributes}

    deepcopy_defined = hasattr(obj, '__deepcopy__')
    if deepcopy_defined:
        # Do hack to prevent infinite recursion in call to deepcopy
        deepcopy_method = obj.__deepcopy__
        obj.__deepcopy__ = None

    for attr in shared_attributes:
        del obj.__dict__[attr]

    clone = deepcopy(obj, memo)

    for attr, val in shared_attrs.items():
        setattr(obj, attr, val)
        setattr(clone, attr, val)

    if deepcopy_defined:
        # Undo hack
        obj.__deepcopy__ = deepcopy_method
        del clone.__deepcopy__

    return clone


def get_uuid(obj: str | UUID | objs.ParentRef | objs.NotionObject) -> UUID:
    """Retrieves a UUID from an object reference.

    Only meant for internal use.
    """
    return objs.ObjectReference.build(obj).id


KT = TypeVar('KT')
VT = TypeVar('VT')


def dict_diff(dct1: dict[KT, VT], dct2: dict[KT, VT]) -> tuple[list[KT], list[KT], dict[KT, tuple[VT, VT]]]:
    """Returns the added keys, removed keys and keys of changed values of both dictionaries."""
    set1, set2 = set(dct1.keys()), set(dct2.keys())
    keys_added = list(set2 - set1)
    keys_removed = list(set1 - set2)
    values_changed = {key: (dct1[key], dct2[key]) for key in set1 & set2 if dct1[key] != dct2[key]}
    return keys_added, keys_removed, values_changed


def dict_diff_str(dct1: dict[KT, VT], dct2: dict[KT, VT]) -> tuple[str, str, str]:
    """Returns the added keys, removed keys and keys of changed values of both dictionaries as strings for printing."""
    keys_added, keys_removed, values_changed = dict_diff(dct1, dct2)
    keys_added_str = ', '.join([str(k) for k in keys_added]) or 'None'
    keys_removed_str = ', '.join([str(k) for k in keys_removed]) or 'None'
    keys_changed_str = ', '.join(f'{k}: {v[0]} -> {v[1]}' for k, v in values_changed.items()) or 'None'
    return keys_added_str, keys_removed_str, keys_changed_str


Self = TypeVar('Self', bound='Wrapper[Any]')  # ToDo: Replace when requires-python >= 3.11
GT = TypeVar('GT', bound=GenericObject)  # ToDo: Use new syntax when requires-python >= 3.12


class Wrapper(Generic[GT]):
    """Convert objects from the obj-based API to the high-level API and vice versa."""

    obj_ref: GT

    _obj_api_map: ClassVar[dict[type[GT], type[Wrapper]]] = {}  # type: ignore[misc]

    def __init_subclass__(cls, wraps: type[GT], **kwargs: Any):
        super().__init_subclass__(**kwargs)
        cls._obj_api_map[wraps] = cls

    def __new__(cls: type[Wrapper], *args, **kwargs) -> Wrapper:
        # Needed for wrap_obj_ref and its call to __new__ to work!
        return super().__new__(cls)

    def __init__(self, *args: Any, **kwargs: Any):
        """Default constructor that also builds `obj_ref`."""
        obj_api_type: type[GenericObject] = self._obj_api_map_inv[self.__class__]
        self.obj_ref = obj_api_type.build(*args, **kwargs)

    @classmethod
    def wrap_obj_ref(cls: type[Self], obj_ref: GT, /) -> Self:
        """Wraps `obj_ref` into a high-level object for the API of Ultimate Notion."""
        hl_cls = cls._obj_api_map[type(obj_ref)]
        hl_obj = hl_cls.__new__(hl_cls)
        hl_obj.obj_ref = obj_ref
        return cast(Self, hl_obj)

    @property
    def _obj_api_map_inv(self) -> dict[type[Wrapper], type[GT]]:
        return {v: k for k, v in self._obj_api_map.items()}


def get_active_session() -> Session:
    """Return the current active session or raise an exception.

    Avoids cyclic imports when used within the package itself.
    For internal use mostly.
    """
    from ultimate_notion.session import Session  # noqa: PLC0415

    return Session.get_active()


def get_repr(obj: Any, /, *, name: Any = None, desc: Any = None) -> str:
    """Default representation, i.e. `repr(...)`, used by us for consistency."""
    type_str = str(name) if name is not None else obj.__class__.__name__
    desc_str = str(desc) if desc is not None else str(obj)
    return f"<{type_str}: '{desc_str}' at {hex(id(obj))}>"


def convert_md_to_py(path: Path | str):
    """Converts a Markdown file to a py file by extracting all python codeblocks"""
    if isinstance(path, str):
        path = Path(path)
    if not path.is_file():
        msg = f'{path} is no file!'
        raise RuntimeError(msg)

    md_str = path.read_text()

    def check_codeblock(block):
        first_line = block.split('\n')[0]
        if first_line[3:] != 'python':
            return ''
        return '\n'.join(block.split('\n')[1:])

    docstring = textwrap.dedent(md_str)
    in_block = False
    block = ''
    codeblocks = []
    for line in docstring.split('\n'):
        if line.startswith('```'):
            if in_block:
                codeblocks.append(check_codeblock(block))
                block = ''
            in_block = not in_block
        if in_block:
            block += line + '\n'
    py_str = '\n'.join([c for c in codeblocks if c != ''])

    path.with_suffix('.py').write_text(py_str)



"""Iterator classes for working with paginated API responses."""

from __future__ import annotations

import logging
from typing import Annotated, Any

from pydantic import Field
from pydantic.functional_validators import BeforeValidator

from ultimate_notion.obj_api.blocks import Block, Database, Page
from ultimate_notion.obj_api.core import GenericObject, NotionObject, TypedObject
from ultimate_notion.obj_api.objects import User
from ultimate_notion.obj_api.props import PropertyItem

MAX_PAGE_SIZE = 100

logger = logging.getLogger(__name__)


def convert_to_notion_obj(obj: dict[str, Any]) -> Block | Page | Database | PropertyItem | User | GenericObject:
    """Convert a dictionary to the corresponding subtype of Notion Object.

    Used in the ObjectList below the convert the results from the Notion API.
    """

    if 'object' not in obj:
        msg = 'Unknown object in results'
        raise ValueError(msg)

    if obj['object'] == BlockList.build().type:  # .build() as the model is not constructed at that point.
        return Block.model_validate(obj)

    if obj['object'] == PageList.build().type:
        return Page.model_validate(obj)

    if obj['object'] == DatabaseList.build().type:
        return Database.model_validate(obj)

    if obj['object'] == PropertyItemList.build().type:
        return PropertyItem.model_validate(obj)

    if obj['object'] == UserList.build().type:
        return User.model_validate(obj)

    return GenericObject.model_validate(obj)


class ObjectList(NotionObject, TypedObject, object='list', polymorphic_base=True):
    """A paginated list of objects returned by the Notion API.

    More details in the [Notion API](https://developers.notion.com/reference/intro#responses).
    """

    results: list[Annotated[NotionObject, BeforeValidator(convert_to_notion_obj)]] = Field(default_factory=list)
    has_more: bool = False
    next_cursor: str | None = None


class BlockList(ObjectList, type='block'):
    """A list of Block objects returned by the Notion API."""

    block: Any = Field(default_factory=dict)


class PageList(ObjectList, type='page'):
    """A list of Page objects returned by the Notion API."""

    page: Any = Field(default_factory=dict)


class DatabaseList(ObjectList, type='database'):
    """A list of Database objects returned by the Notion API."""

    database: Any = Field(default_factory=dict)


class PageOrDatabaseList(ObjectList, type='page_or_database'):
    """A list of Page or Database objects returned by the Notion API."""

    page_or_database: Any = Field(default_factory=dict)


class UserList(ObjectList, type='user'):
    """A list of User objects returned by the Notion API."""

    user: Any = Field(default_factory=dict)


class PropertyItemList(ObjectList, type='property_item'):
    """A paginated list of property items returned by the Notion API.

    Property item lists contain one or more pages of basic property items.  These types
    do not typically match the schema for corresponding property values.
    """

    class _NestedData(GenericObject):
        id: str = None  # type: ignore  # noqa: A003
        type: str = None  # type: ignore  # noqa: A003
        next_url: str | None = None

    property_item: _NestedData = _NestedData()


class EndpointIterator:
    """Iterates over results from a paginated API response.

    These objects may be reused, however they are not thread safe. For example,
    after creating the following iterator:

        notion = notional.connect(auth=NOTION_TOKEN)
        query = EndpointIterator(notion.databases().query)

    The iterator may be reused with different database ID's:

        for items in query(database_id=first_db):
            ...

        for items in query(database_id=second_db):
            ...

    Objects returned by the iterator may also be converted to a specific type.  This
    is most commonly used to wrap API objects with a higher-level object (such as ORM
    types).
    """

    def __init__(self, endpoint):
        """Initialize an object list iterator for the specified endpoint.

        If a class is provided, it will be constructued for each result returned by
        this iterator.  The constructor must accept a single argument, which is the
        `NotionObject` contained in the `ObjectList`.
        """
        self._endpoint = endpoint

        self.has_more = None
        self.page_num = -1
        self.total_items = -1
        self.next_cursor = None

    def __call__(self, **kwargs):
        """Return a generator for this endpoint using the given parameters."""

        self.has_more = True
        self.page_num = 0
        self.total_items = 0

        if 'page_size' not in kwargs:
            kwargs['page_size'] = MAX_PAGE_SIZE

        self.next_cursor = kwargs.pop('start_cursor', None)

        while self.has_more:
            self.page_num += 1

            result_page = self._endpoint(start_cursor=self.next_cursor, **kwargs)

            api_list = ObjectList.model_validate(result_page)

            for obj in api_list.results:
                self.total_items += 1
                yield obj

            self.next_cursor = api_list.next_cursor
            self.has_more = api_list.has_more and self.next_cursor is not None

    def as_list(self, **kwargs):
        """Collect all items from the endpoint as a list."""

        items = []

        for item in self(**kwargs):
            items.append(item)

        return items



"""Utilities for working text, markdown & rich text in Notion."""

from __future__ import annotations

import re
from collections.abc import Iterator

#: the max text size according to the Notion API is 2000 characters.
MAX_TEXT_OBJECT_SIZE = 2000

BASE_URL_PATTERN = r'https://(www.)?notion.so/'
UUID_PATTERN = r'[0-9a-f]{8}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{12}'

UUID_RE = re.compile(rf'^(?P<id>{UUID_PATTERN})$')

PAGE_URL_SHORT_RE = re.compile(
    rf"""^
      {BASE_URL_PATTERN}
      (?P<page_id>{UUID_PATTERN})
    $""",
    flags=re.IGNORECASE | re.VERBOSE,
)

PAGE_URL_LONG_RE = re.compile(
    rf"""^
      {BASE_URL_PATTERN}
      (?P<title>.*)-
      (?P<page_id>{UUID_PATTERN})
    $""",
    flags=re.IGNORECASE | re.VERBOSE,
)

BLOCK_URL_LONG_RE = re.compile(
    rf"""^
      {BASE_URL_PATTERN}
      (?P<username>.*)/
      (?P<title>.*)-
      (?P<page_id>{UUID_PATTERN})
      \#(?P<block_id>{UUID_PATTERN})
    $""",
    flags=re.IGNORECASE | re.VERBOSE,
)


def extract_id(text: str) -> str | None:
    """Examine the given text to find a valid Notion object ID."""

    m = UUID_RE.match(text)
    if m is not None:
        return m.group('id')

    m = PAGE_URL_LONG_RE.match(text)
    if m is not None:
        return m.group('page_id')

    m = PAGE_URL_SHORT_RE.match(text)
    if m is not None:
        return m.group('page_id')

    m = BLOCK_URL_LONG_RE.match(text)
    if m is not None:
        return m.group('block_id')

    return None


def chunky(text: str, length: int = MAX_TEXT_OBJECT_SIZE) -> Iterator[str]:
    """Break the given `text` into chunks of at most `length` size."""
    return (text[idx : idx + length] for idx in range(0, len(text), length))


def python_identifier(string: str) -> str:
    """Make a valid Python identifier.

    This will remove any leading characters that are not valid and change all
    invalid interior sequences to underscore.

    Attention: This may result in an empty string!
    """
    s = re.sub(r'[^0-9a-zA-Z_]+', '_', string)
    s = re.sub(r'^[^a-zA-Z]+', '', s)
    return s.rstrip('_')


def snake_case(string: str) -> str:
    """Make a Python identifier in snake_case.

    Attention: This may result in an empty string!
    """
    return python_identifier(string).lower()


def camel_case(string: str) -> str:
    """Make a Python identifier in CamelCase.

    Attention: This may result in an empty string and a CamelCase sting will be capitalized!
    """
    return ''.join([elem.capitalize() for elem in snake_case(string).split('_')])


def decapitalize(string: str) -> str:
    """Inverse of `capitalize`."""
    if not string:
        return string
    return string[0].lower() + string[1:]


def html_img(url: str, size: float) -> str:
    """Create a img tag in HTML."""
    return f'<img src="{url}" style="height:{size:.2f}em">'



"""Wrapper for various Notion API objects like parents, mentions, emojis & users

Similar to other records, these object provide access to the primitive data structure
used in the Notion API as well as higher-level methods.

For validation the Pydantic model fields specify if a field is optional or not.
Some fields are always set, e.g. `id`, when retrieving an object but must not be set
when sending the object to the Notion API in order to create the object.
To model this behavior, the default value `None` is used for those objects, e.g.
```
class SelectOption(GenericObject)
    id: str = None  # type: ignore  # to make sure mypy doesn't complain
```
Also be aware that this is import when updating to differentiate actual set values
from default/unset values.
"""

from __future__ import annotations

from abc import ABC
from copy import deepcopy
from datetime import date, datetime
from enum import Enum
from typing import TYPE_CHECKING
from uuid import UUID

from pydantic import Field

from ultimate_notion.obj_api.core import GenericObject, NotionObject, TypedObject
from ultimate_notion.obj_api.enums import BGColor, Color
from ultimate_notion.text import extract_id

if TYPE_CHECKING:
    from ultimate_notion.obj_api.blocks import Database, Page


class SelectOption(GenericObject):
    """Options for select & multi-select objects."""

    name: str
    id: str = None  # type: ignore  # noqa: A003  # According to docs: "These are sometimes, but not always, UUIDs."
    color: Color = Color.DEFAULT
    description: str | None = None  # ToDo: Undocumented, could also be [RichTextObject]

    @classmethod
    def build(cls, name, color=Color.DEFAULT):
        """Create a `SelectOption` object from the given name and color."""
        return cls.model_construct(name=name, color=color)


class SelectGroup(GenericObject):
    """Group of options for status objects."""

    name: str
    id: str = None  # type: ignore  # noqa: A003  # According to docs: "These are sometimes, but not always, UUIDs."
    color: Color = Color.DEFAULT
    option_ids: list[str] = Field(default_factory=list)


class ObjectReference(GenericObject):
    """A general-purpose object reference in the Notion API."""

    id: UUID  # noqa: A003

    @classmethod
    def build(cls, ref):
        """Compose an ObjectReference from the given reference.

        `ref` may be a `UUID`, `str`, `ParentRef` or `GenericObject` with an `id`.

        Strings may be either UUID's or URL's to Notion content.
        """

        if isinstance(ref, cls):
            return ref.model_copy(deep=True)

        if isinstance(ref, ParentRef):
            # ParentRef's are typed-objects with a nested UUID
            return ObjectReference.model_construct(id=ref())

        if isinstance(ref, GenericObject) and hasattr(ref, 'id'):
            # re-compose the ObjectReference from the internal ID
            return ObjectReference.build(ref.id)

        if isinstance(ref, UUID):
            return ObjectReference.model_construct(id=ref)

        if isinstance(ref, str):
            ref = extract_id(ref)

            if ref is not None:
                return ObjectReference.model_construct(id=UUID(ref))

        msg = "Unrecognized 'ref' attribute"
        raise ValueError(msg)


# https://developers.notion.com/reference/parent-object
class ParentRef(TypedObject, ABC, polymorphic_base=True):
    """Reference another block as a parent.

    Note: This class is simply a placeholder for the typed concrete *Ref classes.
          Callers should always instantiate the intended concrete versions.
    """


class DatabaseRef(ParentRef, type='database_id'):
    """Reference a database."""

    database_id: UUID

    @classmethod
    def build(cls, db_ref: Database | str | UUID):
        """Compose a DatabaseRef from the given reference object."""
        ref = ObjectReference.build(db_ref)
        return DatabaseRef.model_construct(database_id=ref.id)


class PageRef(ParentRef, type='page_id'):
    """Reference a page."""

    page_id: UUID

    @classmethod
    def build(cls, page_ref: Page | str | UUID):
        """Compose a PageRef from the given reference object."""
        ref = ObjectReference.build(page_ref)
        return PageRef.model_construct(page_id=ref.id)


class BlockRef(ParentRef, type='block_id'):
    """Reference a block."""

    block_id: UUID

    @classmethod
    def build(cls, block_ref):
        """Compose a BlockRef from the given reference object.

        `block_ref` can be either a string, UUID, or block.
        """
        ref = ObjectReference[block_ref]
        return BlockRef.model_construct(block_id=ref.id)


class WorkspaceRef(ParentRef, type='workspace'):
    """Reference the workspace."""

    workspace: bool = True


class UserRef(NotionObject, object='user'):
    """Reference to a user, e.g. in `created_by`, `last_edited_by`, mentioning, etc."""


class UserType(str, Enum):
    """Available user types."""

    PERSON = 'person'
    BOT = 'bot'


class User(UserRef, TypedObject, polymorphic_base=True):
    """Represents a User in Notion."""

    name: str | None = None
    avatar_url: str | None = None


class Person(User, type='person'):
    """Represents a Person in Notion."""

    class _NestedData(GenericObject):
        email: str = None  # type: ignore

    person: _NestedData = _NestedData()


class Bot(User, type='bot'):
    """Represents a Bot in Notion."""

    class _NestedData(GenericObject):
        owner: WorkspaceRef = None  # type: ignore
        workspace_name: str = None  # type: ignore

    bot: _NestedData = _NestedData()


class EmojiObject(TypedObject, type='emoji'):
    """A Notion emoji object."""

    emoji: str

    @classmethod
    def build(cls, emoji):
        """Compose an EmojiObject from the given emoji string."""
        # Todo: convert string-based emoji to unicode here!
        return EmojiObject.model_construct(emoji=emoji)


class FileObject(TypedObject, polymorphic_base=True):
    """A Notion file object.

    Depending on the context, a FileObject may require a name (such as in the `Files`
    property).  This makes the object hierarchy difficult, so here we simply allow
    `name` to be optional. It is the responsibility of the caller to set `name` if
    required by the API.
    """

    name: str | None = None


class HostedFile(FileObject, type='file'):
    """A Notion file object."""

    class _NestedData(GenericObject):
        url: str
        expiry_time: datetime | None = None

    file: _NestedData


class ExternalFile(FileObject, type='external'):
    """An external file object."""

    class _NestedData(GenericObject):
        url: str

    external: _NestedData

    @classmethod
    def build(cls, url, name=None):
        """Create a new `ExternalFile` from the given URL."""
        return cls.model_construct(name=name, external=cls._NestedData(url=url))


class DateRange(GenericObject):
    """A Notion date range, with an optional end date."""

    start: date | datetime
    end: date | datetime | None = None
    time_zone: str | None = None


class Annotations(GenericObject):
    """Style information for RichTextObject's."""

    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    color: BGColor = None  # type: ignore

    @property
    def is_plain(self):
        """Determine if any flags are set in this `Annotations` object.

        If all flags match their defaults, this is considered a "plain" style.
        """

        # Todo: a better approach here would be to just compare all fields to defaults

        if self.bold:
            return False
        if self.italic:
            return False
        if self.strikethrough:
            return False
        if self.underline:
            return False
        if self.code:
            return False
        if self.color is not None:
            return False
        return True


class RichTextObject(TypedObject, polymorphic_base=True):
    """Base class for Notion rich text elements."""

    plain_text: str
    href: str | None = None
    annotations: Annotations | None = None

    # def __str__(self):
    #     """Return a string representation of this object."""

    #     if self.href is None:
    #         text = self.plain_text or ""
    #     elif self.plain_text is None or len(self.plain_text) == 0:
    #         text = f"({self.href})"
    #     else:
    #         text = f"[{self.plain_text}]({self.href})"

    #     if self.annotations:
    #         if self.annotations.bold:
    #             text = f"*{text}*"
    #         if self.annotations.italic:
    #             text = f"**{text}**"
    #         if self.annotations.underline:
    #             text = f"_{text}_"
    #         if self.annotations.strikethrough:
    #             text = f"~{text}~"
    #         if self.annotations.code:
    #             text = f"`{text}`"

    #     return text

    @classmethod
    def build(cls, text: str, href: str | None = None, style: Annotations | None = None):
        """Compose a TextObject from the given properties.

        Args:
            text: the plain text of this object
            href: an optional link for this object
            style: an optional Annotations object for this text
        """
        style = deepcopy(style)
        return cls.model_construct(plain_text=text, href=href, annotations=style)


class LinkObject(GenericObject):
    """Reference a URL."""

    type: str = 'url'  # noqa: A003
    url: str = None  # type: ignore


class TextObject(RichTextObject, type='text'):
    """Notion text element."""

    class _NestedData(GenericObject):
        content: str = None  # type: ignore
        link: LinkObject | None = None

    text: _NestedData = _NestedData()

    @classmethod
    def build(cls, text: str, href: str | None = None, style: Annotations | None = None):
        """Compose a TextObject from the given properties.

        Args:
            text: the plain text of this object
            href: optional link for this object
            style: optional annotations for styling this text
        """

        if text is None:
            return None

        link = LinkObject(url=href) if href else None
        nested = TextObject._NestedData(content=text, link=link)
        style = deepcopy(style)

        return cls.model_construct(
            plain_text=text,
            text=nested,
            href=href,
            annotations=style,
        )


class EquationObject(RichTextObject, type='equation'):
    """Notion equation element."""

    class _NestedData(GenericObject):
        expression: str

    equation: _NestedData


class MentionData(TypedObject, polymorphic_base=True):
    """Base class for typed `Mention` data objects."""


class MentionObject(RichTextObject, type='mention'):
    """Notion mention element."""

    mention: MentionData


class MentionUser(MentionData, type='user'):
    """Nested user data for `Mention` properties."""

    user: UserRef

    @classmethod
    def build(cls, user: User) -> MentionObject:
        """Build a `Mention` object for the specified user.

        The `id` field must be set for the given User.  Other fields may cause errors
        if they do not match the specific type returned from the API.
        """
        mention = MentionUser.model_construct(user=user)
        return MentionObject.model_construct(plain_text=str(user), mention=mention)


class MentionPage(MentionData, type='page'):
    """Nested page data for `Mention` properties."""

    page: ObjectReference

    @classmethod
    def build(cls, page_ref) -> MentionObject:
        """Build a `Mention` object for the specified page reference."""

        ref = ObjectReference.build(page_ref)
        mention = MentionPage.model_construct(page=ref)
        return MentionObject.model_construct(plain_text=str(ref), mention=mention)


class MentionDatabase(MentionData, type='database'):
    """Nested database information for `Mention` properties."""

    database: ObjectReference

    @classmethod
    def build(cls, page) -> MentionObject:
        """Build a `Mention` object for the specified database reference."""

        ref = ObjectReference.build(page)
        mention = MentionDatabase.model_construct(database=ref)
        return MentionObject.model_construct(plain_text=str(ref), mention=mention)


class MentionDate(MentionData, type='date'):
    """Nested date data for `Mention` properties."""

    date: DateRange

    @classmethod
    def build(cls, start, end=None) -> MentionObject:
        """Build a `Mention` object for the specified URL."""

        date_obj = DateRange(start=start, end=end)
        mention = MentionDate.model_construct(date=date_obj)
        return MentionObject.model_construct(plain_text=str(date_obj), mention=mention)


class MentionLinkPreview(MentionData, type='link_preview'):
    """Nested url data for `Mention` properties.

    These objects cannot be created via the API.
    """

    class _NestedData(GenericObject):
        url: str

    link_preview: _NestedData


class MentionTemplateData(TypedObject):
    """Nested template data for `Mention` properties."""


class MentionTemplate(MentionData, type='template_mention'):
    """Nested template data for `Mention` properties."""

    template_mention: MentionTemplateData


class MentionTemplateDate(MentionTemplateData, type='template_mention_date'):
    """Nested date template data for `Mention` properties."""

    template_mention_date: str


class MentionTemplateUser(MentionTemplateData, type='template_mention_user'):
    """Nested user template data for `Mention` properties."""

    template_mention_user: str



"""Wrapper for Notion API blocks.

Blocks are the base for all Notion content.
"""

# ToDo: Clean up all the comments in here

from __future__ import annotations

from abc import ABC
from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic import SerializeAsAny

from ultimate_notion.obj_api.core import GenericObject, NotionObject, TypedObject
from ultimate_notion.obj_api.enums import BGColor, CodeLang, Color
from ultimate_notion.obj_api.objects import (
    BlockRef,
    EmojiObject,
    FileObject,
    ParentRef,
    RichTextObject,
    UserRef,
)
from ultimate_notion.obj_api.props import PropertyValue
from ultimate_notion.obj_api.schema import PropertyType


class DataObject(NotionObject):
    """The base type for all Notion API records."""

    id: UUID = None  # type: ignore  # noqa: A003

    parent: SerializeAsAny[ParentRef] = None  # type: ignore
    has_children: bool = False

    archived: bool = False

    created_time: datetime = None  # type: ignore
    created_by: UserRef = None  # type: ignore

    last_edited_time: datetime = None  # type: ignore
    last_edited_by: UserRef = None  # type: ignore


class Database(DataObject, object='database'):
    """A database record type."""

    title: list[RichTextObject] = None  # type: ignore
    url: str = None  # type: ignore
    public_url: str | None = None
    icon: SerializeAsAny[FileObject] | EmojiObject | None = None
    cover: FileObject | None = None
    properties: dict[str, PropertyType] = None  # type: ignore
    description: list[RichTextObject] = None  # type: ignore
    is_inline: bool = False


class Page(DataObject, object='page'):
    """A standard Notion page object."""

    url: str = None  # type: ignore
    public_url: str | None = None
    icon: SerializeAsAny[FileObject] | EmojiObject | None = None
    cover: SerializeAsAny[FileObject] | None = None
    properties: dict[str, PropertyValue] = None  # type: ignore


class Block(DataObject, TypedObject, object='block', polymorphic_base=True):
    """A standard block object in Notion.

    Calling the block will expose the nested data in the object.
    """


class UnsupportedBlock(Block, type='unsupported'):
    """A placeholder for unsupported blocks in the API."""

    class _NestedData(GenericObject):
        pass

    unsupported: _NestedData | None = None


class TextBlock(Block, ABC):
    """A standard abstract text block object in Notion."""

    # text blocks have a nested object with 'type' name and a 'rich_text' child

    # @property
    # def __text__(self):
    #     """Provide shorthand access to the nested text content in this block."""

    #     return self("rich_text")

    # @classmethod
    # def build(cls, *text):
    #     """Compose a `TextBlock` from the given text items."""

    #     obj = cls()
    #     obj.concat(*text)

    #     return obj

    # def concat(self, *text):
    #     """Concatenate text (either `RichTextObject` or `str` items) to this block."""

    #     rtf = rich_text(*text)

    #     # calling the block returns the nested data...  this helps deal with
    #     # sublcasses of `TextBlock` that each have different "type" attributes
    #     nested = self()
    #     nested.rich_text.extend(rtf)

    # @property
    # def PlainText(self):
    #     """Return the contents of this Block as plain text."""

    #     content = self.__text__

    #     return None if content is None else plain_text(*content)


class Paragraph(TextBlock, type='paragraph'):
    """A paragraph block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        children: list[Block] | None = None
        color: Color | BGColor = Color.DEFAULT

    paragraph: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.paragraph and self.paragraph.rich_text:
    #         return markdown(*self.paragraph.rich_text)

    #     return ""


class Heading1(TextBlock, type='heading_1'):
    """A heading_1 block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        color: Color | BGColor = Color.DEFAULT

    heading_1: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.heading_1 and self.heading_1.rich_text:
    #         return f"# {markdown(*self.heading_1.rich_text)} #"

    #     return ""


class Heading2(TextBlock, type='heading_2'):
    """A heading_2 block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        color: Color | BGColor = Color.DEFAULT

    heading_2: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.heading_2 and self.heading_2.rich_text:
    #         return f"## {markdown(*self.heading_2.rich_text)} ##"

    #     return ""


class Heading3(TextBlock, type='heading_3'):
    """A heading_3 block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        color: Color | BGColor = Color.DEFAULT

    heading_3: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.heading_3 and self.heading_3.rich_text:
    #         return f"### {markdown(*self.heading_3.rich_text)} ###"

    #     return ""


class Quote(TextBlock, type='quote'):
    """A quote block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        children: list[Block] | None = None
        color: Color | Color = Color.DEFAULT

    quote: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.quote and self.quote.rich_text:
    #         return "> " + markdown(*self.quote.rich_text)

    #     return ""


class Code(TextBlock, type='code'):
    """A code block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        caption: list[RichTextObject] = None  # type: ignore
        language: CodeLang = CodeLang.PLAIN_TEXT

    code: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, text, lang=CodingLanguage.PLAIN_TEXT):
    #     """Compose a `Code` block from the given text and language."""
    #     block = super().build(text)
    #     block.code.language = lang
    #     return block

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     lang = self.code.language.value if self.code and self.code.language else ""

    #     # FIXME this is not the standard way to represent code blocks in markdown...

    #     if self.code and self.code.rich_text:
    #         return f"```{lang}\n{markdown(*self.code.rich_text)}\n```"

    #     return ""


class Callout(TextBlock, type='callout'):
    """A callout block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        children: list[Block] | None = None
        icon: SerializeAsAny[FileObject] | EmojiObject | None = None
        color: Color | BGColor = BGColor.GRAY

    callout: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, text, emoji=None, color=FullColor.GRAY_BACKGROUND):
    #     """Compose a `Callout` block from the given text, emoji and color."""

    #     if emoji is not None:
    #         emoji = EmojiObject[emoji]

    #     nested = Callout._NestedData(icon=emoji, color=color)

    #     callout = cls(callout=nested)
    #     callout.concat(text)

    #     return callout


class BulletedListItem(TextBlock, type='bulleted_list_item'):
    """A bulleted list item in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        children: list[Block] | None = None
        color: Color | BGColor = Color.DEFAULT

    bulleted_list_item: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.bulleted_list_item and self.bulleted_list_item.rich_text:
    #         return f"- {markdown(*self.bulleted_list_item.rich_text)}"

    #     return ""


class NumberedListItem(TextBlock, type='numbered_list_item'):
    """A numbered list item in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        children: list[Block] | None = None
        color: Color | BGColor = Color.DEFAULT

    numbered_list_item: _NestedData = _NestedData()

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.numbered_list_item and self.numbered_list_item.rich_text:
    #         return f"1. {markdown(*self.numbered_list_item.rich_text)}"

    #     return ""


class ToDo(TextBlock, type='to_do'):
    """A todo list item in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        checked: bool = False
        children: list[Block] | None = None
        color: Color | BGColor = Color.DEFAULT

    to_do: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, text, checked=False, href=None):
    #     """Compose a ToDo block from the given text and checked state."""
    #     return ToDo(
    #         to_do=ToDo._NestedData(
    #             rich_text=[TextObject[text, href]],
    #             checked=checked,
    #         )
    #     )

    # @property
    # def IsChecked(self):
    #     """Determine if this ToDo is marked as checked or not.

    #     If the block is empty (e.g. no nested data), this method returns `None`.
    #     """
    #     return self.to_do.checked if self.to_do else None

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.to_do and self.to_do.rich_text:
    #         if self.to_do.checked:
    #             return f"- [x] {markdown(*self.to_do.rich_text)}"

    #         return f"- [ ] {markdown(*self.to_do.rich_text)}"

    #     return ""


class Toggle(TextBlock, type='toggle'):
    """A toggle list item in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] = None  # type: ignore
        children: list[Block] | None = None
        color: Color | BGColor = Color.DEFAULT

    toggle: _NestedData = _NestedData()


class Divider(Block, type='divider'):
    """A divider block in Notion."""

    divider: Any = None

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""
    #     return "---"


class TableOfContents(Block, type='table_of_contents'):
    """A table_of_contents block in Notion."""

    class _NestedData(GenericObject):
        color: Color | BGColor = Color.DEFAULT

    table_of_contents: _NestedData = _NestedData()


class Breadcrumb(Block, type='breadcrumb'):
    """A breadcrumb block in Notion."""

    class _NestedData(GenericObject):
        pass

    breadcrumb: _NestedData = _NestedData()


class Embed(Block, type='embed'):
    """An embed block in Notion."""

    class _NestedData(GenericObject):
        url: str = None  # type: ignore

    embed: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, url):
    #     """Create a new `Embed` block from the given URL."""
    #     return Embed(embed=Embed._NestedData(url=url))

    # @property
    # def URL(self):
    #     """Return the URL contained in this `Embed` block."""
    #     return self.embed.url

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.embed and self.embed.url:
    #         return f"<{self.embed.url}>"

    #     return ""


class Bookmark(Block, type='bookmark'):
    """A bookmark block in Notion."""

    class _NestedData(GenericObject):
        url: str = None  # type: ignore
        caption: list[RichTextObject] | None = None

    bookmark: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, url):
    #     """Compose a new `Bookmark` block from a specific URL."""
    #     return Bookmark(bookmark=Bookmark._NestedData(url=url))

    # @property
    # def URL(self):
    #     """Return the URL contained in this `Bookmark` block."""
    #     return self.bookmark.url

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.bookmark and self.bookmark.url:
    #         return f"<{self.bookmark.url}>"

    #     return ""


class LinkPreview(Block, type='link_preview'):
    """A link_preview block in Notion."""

    class _NestedData(GenericObject):
        url: str = None  # type: ignore

    link_preview: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, url):
    #     """Create a new `LinkPreview` block from the given URL."""
    #     return LinkPreview(link_preview=LinkPreview._NestedData(url=url))

    # @property
    # def URL(self):
    #     """Return the URL contained in this `LinkPreview` block."""
    #     return self.link_preview.url

    # @property
    # def Markdown(self):
    #     """Return the contents of this block as markdown text."""

    #     if self.link_preview and self.link_preview.url:
    #         return f"<{self.link_preview.url}>"

    #     return ""


class Equation(Block, type='equation'):
    """An equation block in Notion."""

    class _NestedData(GenericObject):
        expression: str = None  # type: ignore

    equation: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, expr):
    #     """Create a new `Equation` block from the given expression."""
    #     return LinkPreview(equation=Equation._NestedData(expression=expr))


class File(Block, type='file'):
    """A file block in Notion."""

    file: SerializeAsAny[FileObject] = None  # type: ignore


class Image(Block, type='image'):
    """An image block in Notion."""

    image: SerializeAsAny[FileObject] = None  # type: ignore


class Video(Block, type='video'):
    """A video block in Notion."""

    video: SerializeAsAny[FileObject] = None  # type: ignore


class PDF(Block, type='pdf'):
    """A pdf block in Notion."""

    pdf: SerializeAsAny[FileObject] = None  # type: ignore


class ChildPage(Block, type='child_page'):
    """A child page block in Notion."""

    class _NestedData(GenericObject):
        title: str = None  # type: ignore

    child_page: _NestedData = _NestedData()


class ChildDatabase(Block, type='child_database'):
    """A child database block in Notion."""

    class _NestedData(GenericObject):
        title: str = None  # type: ignore

    child_database: _NestedData = _NestedData()


class Column(Block, type='column'):
    """A column block in Notion."""

    class _NestedData(GenericObject):
        # note that children will not be populated when getting this block
        # https://developers.notion.com/changelog/column-list-and-column-support
        children: list[Block] | None = None

    column: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, *blocks):
    #     """Create a new `Column` block with the given blocks as children."""
    #     col = cls()

    #     for block in blocks:
    #         col.append(block)

    #     return col


class ColumnList(Block, type='column_list'):
    """A column list block in Notion."""

    class _NestedData(GenericObject):
        # note that children will not be populated when getting this block
        # https://developers.notion.com/changelog/column-list-and-column-support
        children: list[Column] | None = None

    column_list: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, *columns):
    #     """Create a new `Column` block with the given blocks as children."""
    #     cols = cls()

    #     for col in columns:
    #         cols.append(col)

    #     return cols


class TableRow(Block, type='table_row'):
    """A table_row block in Notion."""

    class _NestedData(GenericObject):
        cells: list[list[RichTextObject]] | None = None

        def __getitem__(self, col):
            """Return the cell content for the requested column.

            This will raise an `IndexError` if there are not enough columns.
            """
            if col > len(self.cells):
                raise IndexError()

            return self.cells[col]

    table_row: _NestedData = _NestedData()

    # def __getitem__(self, cell_num):
    #     """Return the cell content for the requested column."""
    #     return self.table_row[cell_num]

    # @classmethod
    # def build(cls, *cells):
    #     """Create a new `TableRow` block with the given cell contents."""
    #     row = cls()

    #     for cell in cells:
    #         row.append(cell)

    #     return row

    # def append(self, text):
    #     """Append the given text as a new cell in this `TableRow`.

    #     `text` may be a string, `RichTextObject` or a list of `RichTextObject`'s.

    #     :param text: the text content to append
    #     """
    #     if self.table_row.cells is None:
    #         self.table_row.cells = []

    #     if isinstance(text, list):
    #         self.table_row.cells.append(list)

    #     elif isinstance(text, RichTextObject):
    #         self.table_row.cells.append([text])

    #     else:
    #         rtf = TextObject[text]
    #         self.table_row.cells.append([rtf])

    # @property
    # def Width(self):
    #     """Return the width (number of cells) in this `TableRow`."""
    #     return len(self.table_row.cells) if self.table_row.cells else 0


class Table(Block, type='table'):
    """A table block in Notion."""

    class _NestedData(GenericObject):
        table_width: int = 0
        has_column_header: bool = False
        has_row_header: bool = False

        # note that children will not be populated when getting this block
        # https://developers.notion.com/reference/block#table-blocks
        children: list[TableRow] | None = None

    table: _NestedData = _NestedData()

    # @classmethod
    # def build(cls, *rows):
    #     """Create a new `Table` block with the given rows."""
    #     table = cls()

    #     for row in rows:
    #         table.append(row)

    #     return table

    # def append(self, block: TableRow):
    #     """Append the given row to this table.

    #     This method is only applicable when creating a new `Table` block.  In order to
    #     add rows to an existing `Table`, use the `blocks.children.append()` endpoint.

    #     When adding a row, this method will rase an exception if the width does not
    #     match the expected number of cells for existing rows in the block.
    #     """

    #     # XXX need to review whether this is applicable during update...  may need
    #     # to raise an error if the block has already been created on the server

    #     if not isinstance(block, TableRow):
    #         raise ValueError("Only TableRow may be appended to Table blocks.")

    #     if self.Width == 0:
    #         self.table.table_width = block.Width
    #     elif self.Width != block.Width:
    #         raise ValueError("Number of cells in row must match table")

    #     self.table.children.append(block)

    # @property
    # def Width(self):
    #     """Return the current width of this table."""
    #     return self.table.table_width


class LinkToPage(Block, type='link_to_page'):
    """A link_to_page block in Notion."""

    link_to_page: SerializeAsAny[ParentRef]


class SyncedBlock(Block, type='synced_block'):
    """A synced_block block in Notion - either original or synced."""

    class _NestedData(GenericObject):
        synced_from: BlockRef | None = None
        children: list[Block] | None = None

    synced_block: _NestedData = _NestedData()


class Template(Block, type='template'):
    """A template block in Notion."""

    class _NestedData(GenericObject):
        rich_text: list[RichTextObject] | None = None
        children: list[Block] | None = None

    template: _NestedData = _NestedData()



"""Functionality around defining a database schema.

Currently only normal databases, no wiki databases, can be created [1].
Neither the `Unique ID` nor `Status` nor the `Verfication` page property can be set as a database column
in a custom Schema when creating the database.

[1] https://developers.notion.com/docs/working-with-databases#wiki-databases


### Design Principles

A schema is a subclass of `PageShema` that holds `Column` objects with a name and an
actual `PropertyType`, e.g. `Text`, `Number`.

The source of truth is always the `obj_ref` and a `PropertyType` holds only auxilliary
information if actually needed. Since the object references `obj_ref` must always point
to the actual `obj_api.blocks.Database.properties` value if the schema is bound to an database,
the method `_remap_obj_refs` rewires this when a schema is used to create a database.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar, cast

from tabulate import tabulate

import ultimate_notion.obj_api.schema as obj_schema
from ultimate_notion.obj_api.schema import AggFunc, NumberFormat
from ultimate_notion.objects import Option, OptionGroup, OptionNS, RichText
from ultimate_notion.props import PropertyValue
from ultimate_notion.text import snake_case
from ultimate_notion.utils import SList, Wrapper, get_active_session, get_repr, is_notebook

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.page import Page

T = TypeVar('T', bound=obj_schema.PropertyType)


class SchemaError(Exception):
    """Raised when there are issues with the schema of a database."""

    def __init__(self, message):
        """Initialize the `SchemaError` with a supplied message."""
        super().__init__(message)


class SchemaNotBoundError(SchemaError):
    """Raised when the schema is not bound to a database."""

    def __init__(self, schema: type[PageSchema]):
        self.schema = schema
        msg = f'Schema {schema.__name__} is not bound to any database'
        super().__init__(msg)


class ReadOnlyColumnError(SchemaError):
    """Raised when a read-only columns tries to be written to."""

    def __init__(self, col: Column):
        self.col = col
        msg = f"Argument {col.attr_name} refers to the read-only column '{col.name}' of type {col.type}"
        super().__init__(msg)


class PageSchema:
    """Base class for the schema of a database."""

    db_title: RichText | None
    db_desc: RichText | None
    _database: Database | None = None

    def __init_subclass__(cls, db_title: RichText | str | None, **kwargs: Any):
        if isinstance(db_title, str):
            db_title = RichText.from_plain_text(db_title)
        cls.db_title = db_title

        cls.db_desc = RichText.from_plain_text(cls.__doc__) if cls.__doc__ is not None else None
        super().__init_subclass__(**kwargs)

    @classmethod
    def from_dict(
        cls, schema_dct: dict[str, PropertyType], db_title: str | None = None, db_desc: str | None = None
    ) -> type[PageSchema]:
        """Creation of a schema from a dictionary for easy support of dynamically created schemas."""
        title_cols = [k for k, v in schema_dct.items() if isinstance(v, Title)]
        if not title_cols:
            msg = 'Missing an item with property type `Title` as value'
            raise SchemaError(msg)
        elif len(title_cols) > 1:
            msg = f'More than one item with property type `Title` as value found: {", ".join(title_cols)}'
            raise SchemaError(msg)

        cls_name = f'{cls.__name__}FromDct'
        attrs: dict[str, Any] = {'db_desc': db_desc}
        for col_name, prop_type in schema_dct.items():
            attrs[snake_case(col_name)] = Column(col_name, prop_type)
        return type(cls_name, (PageSchema,), attrs, db_title=db_title)

    @classmethod
    def create(cls, **kwargs) -> Page:
        """Create a page using this schema with a bound database."""
        return cls.get_db().create_page(**kwargs)

    @classmethod
    def get_cols(cls) -> list[Column]:
        """Get all columns of this schema."""
        return [col for col in cls.__dict__.values() if isinstance(col, Column)]

    @classmethod
    def get_col(cls, col_name: str) -> Column:
        """Get a specific column from this schema assuming that column names are unique."""
        return SList([col for col in cls.get_cols() if col.name == col_name]).item()

    @classmethod
    def to_dict(cls) -> dict[str, PropertyType]:
        """Convert this schema to a dictionary."""
        return {col.name: col.type for col in cls.get_cols()}

    @classmethod
    def as_table(cls, tablefmt: str | None = None) -> str:
        """Return the schema in a given string table format.

        Some table formats:

        - plain: no pseudographics
        - simple: Pandoc's simple table, i.e. only dashes to separate header from content
        - github: GitHub flavored Markdown
        - simple_grid: uses dashes & pipes to separate cells
        - html: standard html markup

        Find more table formats under: https://github.com/astanin/python-tabulate#table-format
        """
        if tablefmt is None:
            tablefmt = 'html' if is_notebook() else 'simple'

        headers = ['Name', 'Property', 'Attribute']
        rows = []
        for col in cls.get_cols():
            rows.append((col.name, col.type, col.attr_name))

        return tabulate(rows, headers=headers, tablefmt=tablefmt)

    @classmethod
    def show(cls, *, simple: bool | None = None):
        """Show the schema as html or as simple table."""
        if simple:
            tablefmt = 'simple'
        elif simple is None:
            tablefmt = 'html' if is_notebook() else 'simple'
        else:
            tablefmt = 'html'

        table_str = cls.as_table(tablefmt)

        if is_notebook() and (tablefmt == 'html'):
            from IPython.display import display_html  # noqa: PLC0415

            display_html(table_str)
        else:
            print(table_str)  # noqa: T201

    @classmethod
    def _repr_html_(cls) -> str:  # noqa: PLW3201
        """Called by Jupyter Lab automatically to display this schema."""
        return cls.as_table(tablefmt='html')

    @classmethod
    def get_title_prop(cls) -> Column:
        """Returns the title property."""
        return SList(col for col in cls.get_cols() if isinstance(col.type, Title)).item()

    @classmethod
    def is_consistent_with(cls, other_schema: type[PageSchema]) -> bool:
        """Is this schema consistent with another ignoring backward relations if not in other schema."""
        own_schema_dct = cls.to_dict()
        other_schema_dct = other_schema.to_dict()

        if own_schema_dct == other_schema_dct:
            # backward relation was initialized in the other schema
            return True

        other_schema_no_backrels_dct = {
            name: prop_type
            for name, prop_type in other_schema_dct.items()
            if not (isinstance(prop_type, Relation) and not prop_type.schema)
        }

        if other_schema_no_backrels_dct == own_schema_dct:
            # backward relation was not yet initialised in the other schema (during the creation of the data model)
            return True

        return False

    @classmethod
    def get_db(cls) -> Database:
        """Get the database that is bound to this schema."""
        if cls.is_bound() and cls._database:
            return cls._database
        else:
            raise SchemaNotBoundError(cls)

    @classmethod
    def bind_db(cls, db: Database):
        """Bind the PageSchema to the corresponding database for back-reference."""
        cls._database = db
        cls._set_obj_refs()

    @classmethod
    def is_bound(cls) -> bool:
        """Determines if the schema is bound to a database."""
        return cls._database is not None

    @classmethod
    def _get_fwd_rels(cls) -> list[Column]:
        return [
            col
            for col in cls.get_cols()
            if isinstance(col.type, Relation) and not (col.type._is_two_way_target or col.type.is_self_ref)
        ]

    @classmethod
    def _init_fwd_rels(cls):
        """Initialise all non-self-referencing forward relations assuming that the target schemas exist."""
        for col in cls._get_fwd_rels():
            col_type = cast(Relation, col.type)
            col_type._make_obj_ref()

    @classmethod
    def _get_self_refs(cls) -> list[Column]:
        """Get all self-referencing relation columns."""
        return [col for col in cls.get_cols() if isinstance(col.type, Relation) and col.type.is_self_ref]

    @classmethod
    def _has_self_refs(cls) -> bool:
        """Determine if self-referencing relation columns are present."""
        return bool([col for col in cls.get_cols() if isinstance(col.type, Relation) and col.type.is_self_ref])

    @classmethod
    def _init_self_refs(cls):
        """Initialise all forward self-referencing relations."""
        if not cls._has_self_refs():
            return
        db = cls.get_db()  # raises if not bound!
        for col in cls._get_self_refs():
            col_type = cast(Relation, col.type)
            col_type._schema = cls  # replace placeholder `SelfRef` with this schema
            col_type._make_obj_ref()

        self_refs_dct = {col.name: col.type.obj_ref for col in cls._get_self_refs()} or None
        session = get_active_session()
        db.obj_ref = session.api.databases.update(db.obj_ref, schema=self_refs_dct)
        cls._set_obj_refs()

    @classmethod
    def _get_init_cols(cls) -> list[Column]:
        """Get all columns that are initialized by now."""
        return [col for col in cls.get_cols() if hasattr(col.type, 'obj_ref')]

    @classmethod
    def _update_bwd_rels(cls):
        """Update the default property name in case of a two-way relation in the external target schema.

        By default the property in the target schema is named "Related to <this_database> (<this_field>)"
        which is then set to the name specified as backward relation.
        """
        for prop_type in cls.to_dict().values():
            if isinstance(prop_type, Relation) and prop_type.is_two_way:
                prop_type._update_bwd_rel()

    @classmethod
    def _set_obj_refs(cls):
        """Map obj_refs from the properties of the schema to obj_ref.properties of the bound database."""
        db_props_dct = cls.get_db().obj_ref.properties
        for prop_name, prop_type in cls.to_dict().items():
            obj_ref = db_props_dct.get(prop_name)
            if obj_ref:
                prop_type.obj_ref = obj_ref


class PropertyType(Wrapper[T], wraps=obj_schema.PropertyType):
    """Base class for Notion property objects.

    Property types define the value types of columns in a database, e.g. number, date, text, etc.
    """

    obj_ref: T
    #: If the Notion API allows to create a new database with a column of this type
    allowed_at_creation = True
    #: Back reference to the column having this type
    col_ref: Column | None = None

    @property
    def id(self) -> str | None:  # noqa: A003
        """Return identifier of this property type."""
        return self.obj_ref.id

    @property
    def name(self) -> str | None:
        """Return name of this property type."""
        return self.obj_ref.name

    @property
    def prop_value(self) -> type[PropertyValue]:
        """Return the corresponding property value of this property type."""
        return PropertyValue._type_value_map[self.obj_ref.type]

    @property
    def readonly(self) -> bool:
        """Return if this property type is read-only."""
        return self.prop_value.readonly

    def __init__(self, *args, **kwargs):
        obj_api_type = self._obj_api_map_inv[self.__class__]
        self.obj_ref = obj_api_type.build(*args, **kwargs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PropertyType):
            return NotImplemented
        return self.obj_ref.type == other.obj_ref.type and self.obj_ref.value == self.obj_ref.value

    def __hash__(self) -> int:
        return hash(self.obj_ref.type) + hash(self.obj_ref.value)

    def __repr__(self) -> str:
        return get_repr(self, name='PropertyType', desc=self.__class__.__name__)

    def __str__(self) -> str:
        return self.__class__.__name__


class Column:
    """Column with a name and a certain Property Type for defining a Notion database schema.

    This is implemented as a descriptor.
    """

    _name: str
    _type: PropertyType
    # properties below are set by __set_name__
    _schema: type[PageSchema]
    _attr_name: str  # Python attribute name of the property in the schema

    def __init__(self, name: str, type: PropertyType) -> None:  # noqa: A002
        self._name = name
        self._type = type

    def __set_name__(self, owner: type[PageSchema], name: str):
        self._schema = owner
        self._attr_name = name
        self._type.col_ref = self  # link back to allow access to _schema, _py_name e.g. for relations

    def __repr__(self) -> str:
        return get_repr(self, name='Column', desc=self.type)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str):
        raise NotImplementedError

    @property
    def type(self) -> PropertyType:  # noqa: A003
        return self._type

    @type.setter
    def type(self, new_type: PropertyType):  # noqa: A003
        raise NotImplementedError

    @property
    def attr_name(self) -> str:
        return self._attr_name


class Title(PropertyType[obj_schema.Title], wraps=obj_schema.Title):
    """Defines the mandatory title column in a database."""


class Text(PropertyType[obj_schema.RichText], wraps=obj_schema.RichText):
    """Defines a text column in a database."""


class Number(PropertyType[obj_schema.Number], wraps=obj_schema.Number):
    """Defines a number column in a database."""

    def __init__(self, number_format: NumberFormat):
        super().__init__(number_format)


class Select(PropertyType[obj_schema.Select], wraps=obj_schema.Select):
    """Defines a select column in a database."""

    def __init__(self, options: list[Option] | type[OptionNS]):
        if isinstance(options, type) and issubclass(options, OptionNS):
            options = options.to_list()

        option_objs = [option.obj_ref for option in options]
        super().__init__(option_objs)

    @property
    def options(self) -> dict[str, Option]:
        return {option.name: Option.wrap_obj_ref(option) for option in self.obj_ref.select.options}


class MultiSelect(PropertyType[obj_schema.MultiSelect], wraps=obj_schema.MultiSelect):
    """Defines a multi-select column in a database."""

    def __init__(self, options: list[Option] | type[OptionNS]):
        if isinstance(options, type) and issubclass(options, OptionNS):
            options = options.to_list()

        option_objs = [option.obj_ref for option in options]
        super().__init__(option_objs)

    @property
    def options(self) -> dict[str, Option]:
        return {option.name: Option.wrap_obj_ref(option) for option in self.obj_ref.multi_select.options}


class Status(PropertyType[obj_schema.Status], wraps=obj_schema.Status):
    """Defines a status column in a database.

    The Notion API doesn't allow to create a column of this type.
    Sending it to the API with options and option groups defined results in an error
    about the existence of the keys `options` and `groups` and removing them
    creates a database with the column missing... ignorance is bliss.

    Also the Status configuration is not mentioned as a
    [Property Schema Object])https://developers.notion.com/reference/property-schema-object).
    """

    allowed_at_creation = False

    @property
    def options(self) -> list[Option]:
        return [Option.wrap_obj_ref(option) for option in self.obj_ref.status.options]

    @property
    def groups(self) -> list[OptionGroup]:
        return [OptionGroup.wrap_obj_ref(group, options=self.options) for group in self.obj_ref.status.groups]


class Date(PropertyType[obj_schema.Date], wraps=obj_schema.Date):
    """Defines a date column in a database."""


class People(PropertyType[obj_schema.People], wraps=obj_schema.People):
    """Defines a people column in a database."""


class Files(PropertyType[obj_schema.Files], wraps=obj_schema.Files):
    """Defines a files column in a database."""


class Checkbox(PropertyType[obj_schema.Checkbox], wraps=obj_schema.Checkbox):
    """Defines a checkbox column in database."""


class Email(PropertyType[obj_schema.Email], wraps=obj_schema.Email):
    """Defines an e-mail column in a database."""


class URL(PropertyType[obj_schema.URL], wraps=obj_schema.URL):
    """Defines a URL column in a database."""


class PhoneNumber(PropertyType[obj_schema.PhoneNumber], wraps=obj_schema.PhoneNumber):
    """Defines a phone number column in a database."""


class Formula(PropertyType[obj_schema.Formula], wraps=obj_schema.Formula):
    """Defines a formula column in a database.

    Currently the formula expression cannot reference other formula columns, e.g. `prop("other formula")`
    This is a limitation of the API.
    """

    def __init__(self, expression: str):
        super().__init__(expression)


class RelationError(SchemaError):
    """Error if a Relation cannot be initialised."""


class SelfRef(PageSchema, db_title=None):
    """Target schema for self-referencing database relations."""


class Relation(PropertyType[obj_schema.Relation], wraps=obj_schema.Relation):
    """Relation to another database."""

    _schema: type[PageSchema] | None = None  # other schema, i.e. of the target database
    _two_way_col: Column | None = None  # other column, i.e. of the target database

    def __init__(self, schema: type[PageSchema] | None = None, *, two_way_col: Column | None = None):
        if two_way_col and not schema:
            msg = '`schema` needs to be provided if `two_way_col` is set'
            raise RuntimeError(msg)

        if isinstance(schema, Column):
            msg = 'Please provide a schema, not a column! Use `two_way_col` to specify a column.'
            raise ValueError(msg)

        self._schema = schema

        if two_way_col is not None:
            if not isinstance(two_way_col.type, Relation):
                msg = f'The two-way column {two_way_col.name} needs to be of type Relation!'
                raise ValueError(msg)

            if two_way_col.type.schema is not None:
                msg = f'The two-way column {two_way_col.name} must not reference a schema itself'
                raise ValueError(msg)

            self._two_way_col = two_way_col

    @property
    def schema(self) -> type[PageSchema] | None:
        """Schema of the relation database."""
        if self._schema:
            return self._schema if self._schema is not SelfRef else None
        elif self.col_ref is not None and self.col_ref._schema.is_bound():
            db = self.col_ref._schema._database
            session = get_active_session()
            return session.get_db(self.obj_ref.relation.database_id).schema if db else None
        else:
            return None

    @property
    def _is_two_way_target(self) -> bool:
        """Determines if relation is meant as target for a two-way relation."""
        return self._schema is None

    @property
    def two_way_col(self) -> Column | None:
        """Return the target column object of a two-way relation."""
        if self._two_way_col:
            return self._two_way_col
        elif (
            hasattr(self, 'obj_ref')
            and self.schema
            and isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation)
        ):
            prop_name = self.obj_ref.relation.dual_property.synced_property_name
            return self.schema.get_col(prop_name) if prop_name else None
        else:
            return None

    @property
    def is_two_way(self) -> bool:
        """Determine if this relation is a two-way relation."""
        return self.two_way_col is not None

    @property
    def is_self_ref(self) -> bool:
        """Determines if this relation is self referencing the same schema."""
        return (self._schema is SelfRef) or (self.col_ref is not None and self._schema is self.col_ref._schema)

    def _make_obj_ref(self):
        """Initialize the low-level object references for this relation.

        This function is used only internally and called during the forward initialization.
        """
        try:
            db = self.schema.get_db()
        except SchemaNotBoundError as e:
            msg = f"A database with schema '{self.schema.__name__}' needs to be created first!"
            raise RelationError(msg) from e

        if self.two_way_col:
            self.obj_ref = obj_schema.DualPropertyRelation.build(db.id)
        else:
            self.obj_ref = obj_schema.SinglePropertyRelation.build(db.id)

    def _update_bwd_rel(self):
        """Change the default name of a two-way relation target to the defined one."""
        if self.col_ref is None:
            msg = 'Trying to inialize a backward relation for one-way relation that is not bound to a column'
            raise SchemaError(msg)

        if not (isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation)):
            msg = f'Trying to inialize backward relation for one-way relation {self.col_ref.name}'
            raise SchemaError(msg)

        obj_synced_property_name = self.obj_ref.relation.dual_property.synced_property_name
        two_way_col_name = self._two_way_col.name
        if obj_synced_property_name != two_way_col_name:
            session = get_active_session()

            # change the old default name in the target schema to what was passed during initialization
            other_db = self.schema.get_db()
            prop_id = self.obj_ref.relation.dual_property.synced_property_id
            schema_dct = {prop_id: obj_schema.RenameProp(name=two_way_col_name)}
            session.api.databases.update(db=other_db.obj_ref, schema=schema_dct)
            other_db.schema._set_obj_refs()

            our_db = self.col_ref._schema.get_db()
            session.api.databases.update(db=our_db.obj_ref, schema={})  # sync obj_ref
            our_db.schema._set_obj_refs()


class RollupError(SchemaError):
    """Error if definition of rollup is wrong."""


class Rollup(PropertyType[obj_schema.Rollup], wraps=obj_schema.Rollup):
    """Defines the rollup column in a database."""

    def __init__(self, relation: Column, property: Column, calculate: AggFunc):  # noqa: A002
        if not isinstance(relation.type, Relation):
            msg = f'Relation {relation} must be of type Relation'
            raise RollupError(msg)
        # ToDo: One could check here if property really is a property in the database where relation points to
        super().__init__(relation.name, property.name, calculate)


class CreatedTime(PropertyType[obj_schema.CreatedTime], wraps=obj_schema.CreatedTime):
    """Defines the created-time column in a database."""


class CreatedBy(PropertyType[obj_schema.CreatedBy], wraps=obj_schema.CreatedBy):
    """Defines the created-by column in a database."""


class LastEditedBy(PropertyType[obj_schema.LastEditedBy], wraps=obj_schema.LastEditedBy):
    """Defines the last-edited-by column in a database."""


class LastEditedTime(PropertyType[obj_schema.LastEditedTime], wraps=obj_schema.LastEditedTime):
    """Defines the last-edited-time column in a database."""


class ID(PropertyType[obj_schema.UniqueID], wraps=obj_schema.UniqueID):
    """Defines a unique ID column in a database."""

    allowed_at_creation = False

    @property
    def prefix(self) -> str:
        opt_prefix = self.obj_ref.unique_id.prefix
        return '' if opt_prefix is None else opt_prefix


class Verification(PropertyType[obj_schema.Verification], wraps=obj_schema.Verification):
    """Defines a unique ID column in a database."""

    allowed_at_creation = False


class DefaultSchema(PageSchema, db_title=None):
    """Default database schema of Notion.

    As inferred by just creating an empty database in the Notion UI.
    """

    name = Column('Name', Title())
    tags = Column('Tags', MultiSelect([]))


class ColType:
    """Namespace class of all columns types for easier access."""

    Title = Title
    Text = Text
    Number = Number
    Select = Select
    MultiSelect = MultiSelect
    Status = Status
    Date = Date
    People = People
    Files = Files
    Checkbox = Checkbox
    Email = Email
    URL = URL
    PhoneNumber = PhoneNumber
    Formula = Formula
    Relation = Relation
    Rollup = Rollup
    CreatedTime = CreatedTime
    CreatedBy = CreatedBy
    LastEditedTime = LastEditedTime
    LastEditedBy = LastEditedBy
    ID = ID
    Verification = Verification



"""Base classes for working with the Notion API."""

from __future__ import annotations

import logging
from typing import Any, ClassVar
from uuid import UUID

from pydantic import BaseModel, ConfigDict, ValidatorFunctionWrapHandler, field_validator, model_validator

logger = logging.getLogger(__name__)


class GenericObject(BaseModel):
    """The base for all API objects.

    As a general convention, data fields in lower case are defined by the
    Notion API.  Properties in Title Case are provided for convenience.
    """

    model_config = ConfigDict(extra='forbid')

    @classmethod
    def _set_field_default(cls, name: str, default: str):
        """Modify the `BaseModel` field information for a specific class instance.

        This is necessary in particular for subclasses that change the default values
        of a model when defined. Notable examples are `TypedObject` and `NotionObject`.

        Args:
            name: the named attribute in the class
            default: the new default value for the named field
        """
        # Rebuild model to avoid UserWarning about shadowing an attribute in parent.
        # More details here: https://github.com/pydantic/pydantic/issues/6966
        field = cls.model_fields.get(name)
        if field is None:
            msg = f'No field of name {name} in {cls.__name__} found!'
            raise ValueError(msg)
        field.default = default
        field.validate_default = False
        cls.model_rebuild(force=True)

    # https://github.com/pydantic/pydantic/discussions/3139
    def update(self, **data):
        """Update the internal attributes with new data."""

        new_obj_dct = self.model_dump()
        new_obj_dct.update(data)
        new_obj = self.model_validate(new_obj_dct)

        for k, v in new_obj.model_dump().items():
            # model_dump(exclude_defaults=True) was used before, but that prevented restoring pages
            logger.debug('updating object data -- %s => %s', k, v)
            setattr(self, k, getattr(new_obj, k))

        return self

    def serialize_for_api(self):
        """Serialize the object for sending it to the Notion API."""
        # TODO: read-only fields should not be sent to the API
        # https://github.com/jheddings/notional/issues/9

        # the API doesn't like "undefined" values...
        return self.model_dump(mode='json', exclude_none=True, by_alias=True)

    @classmethod
    def build(cls, *args, **kwargs):
        """Use the standard constructur to build the instance. Will be overridden for more complex types."""
        return cls(*args, **kwargs)


class NotionObject(GenericObject):
    """A top-level Notion API resource."""

    object: str  # noqa: A003
    request_id: UUID = None  # type: ignore
    id: UUID | str = None  # type: ignore  # noqa: A003

    def __init_subclass__(cls, *, object=None, **kwargs):  # noqa: A002
        super().__init_subclass__(**kwargs)

    @classmethod
    def __pydantic_init_subclass__(cls, *, object=None, **kwargs):  # noqa: A002, PLW3201
        """Update `GenericObject` defaults for the named object.

        Needed since `model_fields` are not available during __init_subclass__
        See: https://github.com/pydantic/pydantic/issues/5369
        """
        super().__pydantic_init_subclass__(object=object, **kwargs)

        if object is not None:  # if None we inherit 'object' from the base class
            cls._set_field_default('object', default=object)

    @field_validator('object', mode='after')
    @classmethod
    def _verify_object_matches_expected(cls, val):
        """Make sure that the deserialized object matches the name in this class."""

        obj_attr = cls.build().object
        if val != obj_attr:
            msg = f'Invalid object for {obj_attr} - {val}'
            raise ValueError(msg)

        return val


class TypedObject(GenericObject):
    """A type-referenced object.

    Many objects in the Notion API follow a standard pattern with a 'type' property
    followed by additional data.  These objects must specify a `type` attribute to
    ensure that the correct object is created.

    For example, this contains a nested 'detail' object:

        data = {
            type: "detail",
            ...
            detail: {
                ...
            }
        }

    Calling the object provides direct access to the data stored in `{type}`.
    """

    type: str  # noqa: A003
    _polymorphic_base: ClassVar[bool] = False

    def __init_subclass__(cls, *, type: str | None = None, polymorphic_base: bool = False, **kwargs):  # noqa: A002
        cls._polymorphic_base = polymorphic_base
        super().__init_subclass__(**kwargs)

    @classmethod
    def __pydantic_init_subclass__(cls, *, type: str | None = None, **kwargs):  # noqa: A002, PLW3201
        """Register the subtypes of the TypedObject subclass.

        Needed since `model_fields` are not available during __init_subclass__.
        See: https://github.com/pydantic/pydantic/issues/5369
        """
        type_name = cls.__name__ if type is None else type
        cls._register_type(type_name)
        super().__pydantic_init_subclass__(**kwargs)

    @classmethod
    def _register_type(cls, name):
        """Register a specific class for the given 'type' name."""

        cls._set_field_default('type', default=name)

        # initialize a _typemap map for each direct child of TypedObject

        # this allows different class trees to have the same 'type' name
        # but point to a different object (e.g. the 'date' type may have
        # different implementations depending where it is used in the API)

        if not hasattr(cls, '_typemap'):
            cls._typemap = {}

        if name in cls._typemap:
            msg = f'Duplicate subtype for class - {name} :: {cls}'
            raise ValueError(msg)

        logger.debug('registered new subtype: %s => %s', name, cls)

        cls._typemap[name] = cls

    @model_validator(mode='wrap')
    @classmethod
    def _resolve_type(cls, value: Any, handler: ValidatorFunctionWrapHandler):
        """Instantiate the correct object based on the 'type' field.

        Following this approach: https://github.com/pydantic/pydantic/discussions/7008
        Also the reason for `polymorphic_base` is explained there.
        """

        if isinstance(value, cls):
            return handler(value)

        if not cls._polymorphic_base:  # breaks the recursion
            return handler(value)

        if not isinstance(value, dict):
            msg = "Invalid 'data' object"
            raise ValueError(msg)

        if not hasattr(cls, '_typemap'):
            msg = f"Missing '_typemap' in {cls}"
            raise ValueError(msg)

        type_name = value.get('type')

        if type_name is None:
            msg = "Missing 'type' in data"
            raise ValueError(msg)

        sub_cls = cls._typemap.get(type_name)

        if sub_cls is None:
            msg = f'Unsupported sub-type: {type_name}'
            raise ValueError(msg)

        logger.debug('initializing typed object %s :: %s => %s -- %s', cls, type_name, sub_cls, value)

        return sub_cls(**value)

    # ToDo: Check if we shouldn't make this rather abstract
    @property
    def value(self) -> Any:
        """Return the nested object."""
        return getattr(self, self.type)
